import copy

from exlab.interface.serializer import Serializable


class Result(Serializable):
    def __init__(self, config):
        self.config = config

        self.reachedGoal = None
        self.reachedGoalDistance = None
        self.reachedGoalStatus = None
        self.reachedContext = None
        self.reachedContextDistance = None
        self.reachedContextStatus = None
        self.action = None

        self.randomProbability = None
        self.planningSuccess = None
        self.planningDistance = None
        self.planningSteps = None
        self.planningChangeContextSteps = []
        self.performerReplanning = 0
        self.performerReplanningSteps = 0
        self.performerDistance = None
        self.performerDerive = []
    
    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['reachedGoal', 'reachedGoalDistance', 'reachedGoalStatus', 'reachedContext', 'reachedContextDistance',
                                            'reachedContextStatus', 'action', 'randomProbability', 'planningSuccess', 'planningDistance',
                                            'planningSteps', 'planningChangeContextSteps', 'performerReplanning', 'performerReplanning', 'performerReplanningSteps',
                                            'performerDistance', 'performerDerive'])
        return dict_

    # @classmethod
    # def _deserialize(cls, dict_, serializer, obj=None):
    #     if obj is None:
    #         obj = cls()
    #     return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        for attr in ['reachedGoal', 'reachedGoalDistance', 'reachedGoalStatus', 'reachedContext', 'reachedContextDistance', 'reachedContextStatus',
                     'action', 'randomProbability', 'planningSuccess', 'planningDistance',
                     'planningSteps', 'planningChangeContextSteps', 'performerReplanning', 'performerReplanning', 'performerReplanningSteps',
                     'performerDistance', 'performerDerive']:
            if attr in dict_:
                setattr(self, attr, serializer.deserialize(dict_.get(attr)))

    def clone(self, config):
        new = copy.copy(self)
        new.config = config
        return new
    
    def results(self):
        score = 0.
        txt = ''

        if self.config.goal:
            txt += f'goal {self.config.goal} '
            if not self.reachedGoal:
                txt += f'not attempted'
            elif isinstance(self.reachedGoal, str):
                txt += f'{self.reachedGoal}'
            else:
                difference = (self.config.goal - self.reachedGoal).norm()
                score += difference / self.config.goal.space.maxDistance * 5.
                txt += f'and got {self.reachedGoal}, difference is {difference}'
            txt += f'|   '

        if self.config.goalContext:
            txt += f'context {self.config.goalContext} '
            if not self.reachedContext:
                txt += f'not attempted'
            elif isinstance(self.reachedContext, str):
                txt += f'{self.reachedContext}'
            else:
                difference = (self.config.goalContext - self.reachedContext).norm()
                score += difference / self.config.goalContext.space.maxDistance * 5.
                txt += f'and got {self.reachedContext}, difference is {difference}'
            txt += f'|   '

        valid = 'Ok' if score < 0.1 else 'Error'
        return f'{valid}: {score} ({txt})'

    def __repr__(self):
        attrResults = ['action', 'reachedGoal', 'reachedGoalDistance', 'reachedGoalStatus',
                       'reachedContext', 'reachedContextDistance', 'reachedContextStatus', 'randomProbability',
                       'planningSuccess', 'planningDistance', 'planningSteps',
                       'planningChangeContextSteps', 'performerReplanning',
                       'performerReplanningSteps', 'performerDistance', 'performerDerive']
        results = ', '.join(
            [f'{k}: {getattr(self, k)}' for k in attrResults if getattr(self, k)])
        return results



from .model import ModelLearner
from dinos.agents.tools.strategies.random import RandomStrategy


class RandomLearner(ModelLearner):
    def __init__(self, environment, dataset=None, performer=None, planner=None, options={}):
        super().__init__(environment=environment, dataset=dataset,
                         performer=performer, planner=planner, options=options)
        self.trainStrategies.add(RandomStrategy(self))



from exlab.lab.lab import Lab
from exlab.lab.counter import EpisodeAbsoluteIterationCounter
from exlab.utils.path import basepath


basedir = basepath(__file__)

lab = Lab(basedir, defaults='defaults.yml', help='Simple example')

lab.parameter('infos.dataset').import_key('datasets', 'data.yml')
lab.parameter('learner').import_file(['learners', 'learners2'])

lab.filter('learners/*').parameter('dataset').import_key('datasets', 'data.yml')

lab.load(counter=EpisodeAbsoluteIterationCounter)

def run(exp):
    print(exp.config['learner'])

lab.run(run)



import pygame
import pymunk
from pymunk import Vec2d

from PIL import Image

from dinos.representation.map import FeatureMap, MapProxyEntity


class PyMunkMapProxyEntity(MapProxyEntity):
    def __init__(self, entity, manager=None):
        super().__init__(entity, manager=manager)
        self.copyPhysics(entity)
        self.addToMap()

        # self.handlers = {}
        self.addPropertyHandler('position', self.handlerPosition)

    def copyPhysics(self, entity):
        self.shape = None
        self.body = pymunk.Body(body_type=pymunk.Body.KINEMATIC)
        if hasattr(entity, 'body') and entity.body:
            self.body.position = entity.body.position
        if hasattr(entity, 'shape') and entity.shape:
            self.shape = entity.shape.copy()
            self.shape.body = self.body

    def addToMap(self):
        if self.body and self.shape:
            self.manager.physics.add(self.body, self.shape)

    def deleteFromMap(self):
        if self.body and self.shape:
            self.manager.physics.remove(self.body, self.shape)

    # Handlers
    def addPropertyHandler(self, name, handler):
        property_ = self.propertyItem(name)
        # self.handlers[property_] = handler
        self.manager.addPropertyHandler(property_, handler)

    def handlerPosition(self, value):
        if self.body:
            self.body.position = value


class PyMunkFeatureMap(FeatureMap):
    def __init__(self):
        super().__init__(proxyCls=PyMunkMapProxyEntity)

        self.physics = pymunk.Space()
        self.physics.gravity = (0.0, 0.0)
        self.physics.damping = 0.1

    def image(self):
        self.physics.step(0.0001)
        surface = pygame.Surface((800, 600))
        options = pymunk.pygame_util.DrawOptions(surface)
        pymunk.pygame_util.positive_y_is_up = False
        self.physics.debug_draw(options)

        pil_string_image = pygame.image.tostring(surface, "RGBA", False)
        pil_image = Image.frombytes(
            "RGBA", surface.get_size(), pil_string_image)

        return pil_image



'''
    File name: environment.py
    Author: Alexandre Manoury
    Python Version: 3.6
'''

import sys
import time
import copy
import random
import numpy as np
import threading
from pathlib import Path

from exlab.lab.counter import IterationCounter
from exlab.modular.module import manage
from exlab.interface.graph import Graph
from exlab.interface.serializer import Serializer

from dinos.utils.move import MoveConfig
from dinos.data.space import Space
from dinos.data.spacemanager import SpaceManager

from dinos.evaluation.evaluator import Evaluator

from dinos.representation.entity_manager import EntityManager
from dinos.representation.live_entity import LiveEntity
from dinos.representation.property import Property
from dinos.representation.state import State

from .engines.engine import Engine
from .scene import SceneSetup

# from .entity import Entity, PhysicalEntity
# from .property import Property
# from ..utils.logging import Logger
# from ..utils.serializer import Serializer


# def make(id_, scene=None, params={}):
#     from ..utils.loaders import DataManager
#     return DataManager.makeEnv(id_, scene, params)


class Environment(SpaceManager, EntityManager):
    """
    Describes the environment for the agent in an experiment.
    Root entity of a world
    """

    DESCRIPTION = ''
    VARYING = False
    CLASS_ENDNAME = 'Environment'

    ENGINE = Engine

    sceneClasses = []
    defaultSceneCls = None

    RESOURCE_FOLDERNAME = 'resources'

    def __init__(self, sceneCls=None, options={}):
        """
        """
        SpaceManager.__init__(self, 'environment', storesData=False)
        EntityManager.__init__(self, 'environment', entityCls=LiveEntity)
        # self.name = options.get('name', 'Environment')
        self.experiment = None

        assert(self.__class__.ENGINE is not None)
        self.engine = self.__class__.ENGINE(self, options.get('engine', {}))

        manage(self).attach_counter(IterationCounter())
        self.scheduledActionCounter = 0
        self.lock = threading.Lock()
        self.terminateThreads = False

        self.evaluators = []
        self.evaluationTimesteps = [50, 100]
        self.evaluationInterval = 100

        # Configuration
        self.options = options
        # self.discrete = options.get('discrete', False)

        self.threading = False
        self.timestep = options.get('timestep', 3.0)  # 2 seconds per action
        # self.unitWindow = options.get("window", 5)  # number of unit

        self.timeByIteration = []

        # Entities
        self.physicalObjects = []

        # Scenes
        self.scene = None
        self.sceneClass = self.defaultSceneCls

        if sceneCls is not None and issubclass(sceneCls, SceneSetup):
            self.registerScene(sceneCls)
            self.sceneClass = sceneCls
        elif isinstance(sceneCls, str):
            self.sceneClass = self.findScene(sceneCls)
        elif 'scene' in options:
            self.sceneClass = self.findScene(options['scene'])

        if self.sceneClass:
            self.setupScene(self.sceneClass)
    
    def deserializer(self):
        d = Serializer()
        d.set('environment', self)
        d.set('environment', self, category='spaceManager')
        d.attach_finder('entity', self.findEntity)
        d.attach_finder('property', self.findProperty)
        d.attach_finder('test', self.findTest)
        return d
    
    def findTest(self, name):
        return next((test for test in self.tests if test.name == name), None)

    @property
    def counter(self):
        return manage(self).counter
    
    @property
    def iteration(self):
        return manage(self).counter.t

    def describe(self):
        spacesDescription = [f"{space}: {space.dim} dimension(s)" for space in self.spaces]
        spaces = '\n'.join(spacesDescription)
        return f"Environment '{self.__class__.__name__}':\n{self.__class__.DESCRIPTION}\n\nSpaces available:\n{spaces}"

    # def setup(self, dataset=None):
    #     # self.bindSpaces()
    #     self.computeSpaces()
    #     if dataset:
    #         for space in self.actionExplorationSpaces:
    #             self.dataset.convertSpace(space)

    @property
    def envname(self):
        name = self.__class__.__name__
        if name.endswith(self.CLASS_ENDNAME):
            name = name[:-len(self.CLASS_ENDNAME)]
        return name
    
    def resourcesPath(self, filepath=''):
        return Path(sys.modules[self.__class__.__module__].__file__).parent / self.RESOURCE_FOLDERNAME / filepath

    # Scenes
    @classmethod
    def registerScene(cls, sceneCls, default=False):
        if default or not cls.sceneClasses:
            cls.defaultSceneCls = sceneCls
        if sceneCls not in cls.sceneClasses:
            cls.sceneClasses.append(sceneCls)
            sceneCls.environmentClass = cls

    # def addSceneAndSetup(self, scene, overwrite=True):
    #     self.addScene(scene)
    #     if overwrite or not self.scene:
    #         self.setupScene(scene)

    def setupScene(self, sceneCls):
        if isinstance(sceneCls, str):
            sceneCls = self.findScene(sceneCls)
        if sceneCls not in self.sceneClasses:
            self.registerScene(sceneCls)
        if self.scene is not None:
            self.clear()
        self.scene = sceneCls(self)
        self.scene.setup()
        self.world.activate()
        self.scene.setupTests()

    def findScene(self, nameCls):
        sceneCls = next(
            (s for s in self.sceneClasses if s.__name__ == nameCls), None)
        if sceneCls is None:
            raise Exception(
                f'Scene named \'{nameCls}\' not found for environment {self}')
        return sceneCls

    @property
    def tests(self):
        return self.scene.tests

    def setupIteration(self, config=MoveConfig()):
        with self.lock:
            self.scene.setupIteration(config)

    def setupEpisode(self, config=MoveConfig()):
        with self.lock:
            self.scene.setupEpisode(config)

    def setupPreTest(self, test=None):
        with self.lock:
            self.scene.setupPreTest(test)
    
    def setupPreTestPoint(self, test=None, point=None):
        with self.lock:
            self.scene.setupPreTestPoint(test, point)

    def state(self, dataset=None, featureMap=None):
        return State(self, self.world.observe().flat(), dataset=dataset, featureMap=featureMap)
    
    # Evaluation
    def evaluator(self, agent, create=False):
        evaluator = next((eva for eva in self.evaluators if eva.agent == agent), None)
        if not evaluator and create:
            evaluator = Evaluator(agent, self)
            self.evaluators.append(evaluator)
        return evaluator

    def setupEvaluators(self):
        for agent in self.agents():
            self.evaluator(agent, create=True)
    
    def evaluate(self):
        for agent in self.agents():
            self.evaluator(agent, create=True).evaluate()
    
    def evaluations(self):
        data = ''
        for evaluator in self.evaluators:
            data += f'Evaluator {evaluator.agent}:\n'
            for i, evaluation in evaluator.evaluations.items():
                data += f'    Iteration {i} {evaluation}:\n'
                for result in evaluation.results:
                    data += '        ' + result.details().replace('\n', '\n        ') + '\n'
        return data

    # Entities
    def reset(self):
        self._reset()
        if self.scene:
            self.scene._reset()

    def _reset(self):
        pass

    def hardReset(self):
        self.clear()
        if self.scene:
            self.scene.setup()

    # def removeObject(self, obj):
    #     if obj in self.objs:
    #         self.removeEntity(obj)
    #         self.objs.remove(obj)
    #         obj.env = None
    #         self.removePhysics(obj)

    def save(self):
        # print("Saved")
        pass

    def execute(self, action, actionParameters=[], config=None, agent=None, sync=False):
        if isinstance(action, Space):
            action = action.point(actionParameters)
        elif isinstance(action, Property):
            action = action.space.point(actionParameters)

        with self.lock:
            for p in action.flat():
                effector = p.space.boundProperty
                if effector is None or not effector.controllable():
                    raise Exception(
                        f'{p.space} is not bound to an effector!')
                effector.perform(p)
        
            self.scheduledActionCounter += 1
        if sync and self.threading:
            self.waitNextIteration(agent)
        else:
            self.run(evaluating=config.evaluating if config else False)

        return self.reward(action)
    
    def waitNextIteration(self, agent):
        # iteration = self.counter.t
        # while self.counter.t == iteration:
        if self.terminateThreads:
            raise Exception('ThreadTerminated')
        agent.iterationEvent.wait()
        agent.iterationEvent.clear()
        if self.terminateThreads:
            raise Exception('ThreadTerminated')
    
    def checkTerminated(self):
        if self.terminateThreads:
            raise Exception('ThreadTerminated')

    def waitAllScheduledActionsExecuted(self):
        with self.lock:
            awaitedActions = self.countScheduledActions()
        # print(awaitedActions, self.scheduledActionCounter)
        while self.scheduledActionCounter < awaitedActions:
            time.sleep(0.001)
            with self.lock:
                awaitedActions = self.countScheduledActions()
            # print(awaitedActions, self.scheduledActionCounter)
        return awaitedActions > 0

    def step(self, action, actionParameters=[], config=None):
        reward = self.execute(
            action, actionParameters=actionParameters, config=config)
        return self.world.observe(), reward, self.done()

    def _preIteration(self):
        self.scene._preIteration()

    def reward(self, action):
        self.scene.reward(action)

    def done(self):
        return False
    
    def runScheduled(self, evaluating=False):
        if self.scheduledActionCounter == 0:
            return
        self.run(evaluating=evaluating)

    def run(self, evaluating=False):
        # Will run as long as actions are scheduled
        if self.threading:
            for agent, _ in self.scheduledActions().items():
                agent.syncCounter()
            threads = [threading.Thread(target=m) for m in self.scheduledActions().values()]
            for t in threads:
                t.start()

        # Will count as 1 iteration, even if the duration is different from the timestep setting
        lastTime = time.time()
        try:
            while True:
                if self.threading and not self.waitAllScheduledActionsExecuted():
                    return
                with self.lock:
                    self._preIteration()
                    self.engine.run()
                    for host in self.world.hosts():
                        host.scheduledAction = False

                    self.scheduledActionCounter = 0
                    if not evaluating:
                        manage(self).counter.next_iteration()
                        if self.iteration % 10 == 0:
                            self.logger.info(f'{self.iteration}...')
                    if self.threading:
                        for agent in self.agents():
                            agent.iterationEvent.set()
                if not evaluating:
                    self.timeByIteration.append(time.time() - lastTime)
                    lastTime = time.time()
                if not self.threading:
                    return
        except KeyboardInterrupt:
            if self.threading:
                self.logger.error('KeyboardInterrupt! Killing all threads')
                self.terminateThreads = True
                for agent in self.agents():
                    agent.iterationEvent.set()
                for t in threads:
                    t.join()
                self.terminateThreads = False
                self.logger.info('All threads terminated')

    # Wrappers
    def image(self):
        return self.engine.graphicalEngine.image()
    
    def show(self):
        self.engine.show()

    def hide(self):
        self.engine.hide()
    
    @property
    def gui(self):
        return self.engine.graphicalEngine.gui
    
    def displayGui(self, gui=True):
        self.engine.graphicalEngine.displayGui(gui)
    
    # Visual
    def visualizeTimeByIteration(self, options=None):
        return Graph(options=options).plot(self.timeByIteration)
    
    def visualizeEvaluations(self, options=None):
        g = Graph(options=options)
        for eva in self.evaluators:
            g += eva.visualizeEvaluations()
        return g

    # def _serialize(self, options):
    #     dict_ = SpaceManager._serialize(self, options)
    #     dict_.update(Entity._serialize(self, options))
    #     dict_.update(Serializer.serialize(
    #         self, ['options'], options=options))  # 'discrete',
    #     dict_.update({'scene': self.scene.serialize(options)['id']})
    #     return dict_

    # @classmethod
    # def _deserialize(cls, dict_, options={}, obj=None):
    #     obj = obj if obj else cls(
    #         dict_.get('scene'), options=dict_.get('options', {}))

    #     # SpaceManager._deserialize(dict_, options=options, obj=obj)
    #     Entity._deserialize(dict_, options=options, obj=obj)
    #     return obj

    # @classmethod
    # def deserialize(cls, dict_, options={}, obj=None):
    #     obj = obj if obj else cls(options=dict_)
    #     # spaces = [EnvSpace.deserialize(space, obj) for space in dict_.get('spaces', [])]
    #
    #     # Operations
    #     obj.setup()
    #     obj.testbenchs = dict_['testbench']
    #     return obj

    # def bindSpaces(self):
    #     for property in self.properties():
    #         property.bindSpace(self.spaces)

    # def generateTestbench(self):
    #     testbench = {}
    #     number = 20
    #     print(self.testbenchs)
    #     for tbconfig in self.testbenchs.get('spaces', []):
    #         print(self.space)
    #         space = SpaceManager.space(self, tbconfig['space'])

    #         tbSpace = []

    #         '''for v in np.linspace(minb, maxb, num=number):
    #         point = []
    #         for d, minb, maxb in enumerate(zip(space.bounds['min'], space.bounds['max'])):
    #             for v in np.linspace(minb, maxb, num=number):
    #                 point.append(v)'''
    #         # TODO  id:9
    #         '''tbSpace = np.zeros((space.dim ** number, space.dim))
    #         for d, minb, maxb in enumerate(zip(space.bounds['min'], space.bounds['max'])):
    #             for x in np.linspace(minb, maxb, num=number):
    #                 tbSpace[]

    #         def rec(space, depth=0):
    #             data = []
    #             for x in np.linspace(-50, 50, num=20):
    #                 data.append = np.array([x])
    #             if depth >= space.dim:'''
    #         N = 5
    #         if space.dim == 1:
    #             for x in np.linspace(tbconfig['bounds']['min'][0], tbconfig['bounds']['max'][0], num=N):
    #                 tbSpace.append(np.array([x]))
    #         elif space.dim == 2:
    #             for x in np.linspace(tbconfig['bounds']['min'][0], tbconfig['bounds']['max'][0], num=N):
    #                 for y in np.linspace(tbconfig['bounds']['min'][1], tbconfig['bounds']['max'][1], num=N):
    #                     tbSpace.append(np.array([x, y]))
    #             '''for x in np.linspace(20, 70, num=20):
    #                 for y in np.linspace(-10, 10, num=20):
    #                     tbSpace.append(np.array([x, y]))'''
    #         testbench[space.name] = tbSpace
    #     # print(testbench.keys())
    #     # print("Generated testbench for {}".format(', '.join(testbench.keys())))
    #     Logger.main().info("Testbench generated for {}".format(
    #         ', '.join(list(testbench.keys()))))
    #     return testbench



from exlab.utils.io import parameter


def lazyImport(path, className):
    def importer():
        module = __import__(path, fromlist=[className])
        return getattr(module, className)
    
    return importer


class Mapper(object):
    DEFAULT = {}

    @classmethod
    def register(cls, environmentClassName, featureMapPath, featureMapClassName, dict_=None):
        dict_ = parameter(dict_, cls.DEFAULT)
        dict_[environmentClassName] = lazyImport(featureMapPath, featureMapClassName)
    
    @classmethod
    def get(cls, environmentClass, dict_=None):
        dict_ = parameter(dict_, cls.DEFAULT)
        importer = dict_.get(environmentClass.__name__)
        if not importer:
            return
        return importer()()


Mapper.register('PlaygroundEnvironment',
                'dinos.environments.priors.maps.playground', 'PlaygroundFeatureMap')



import sys
import copy
import math
import random
import numpy as np

from enum import Enum
from scipy.spatial.distance import euclidean
from sklearn.neighbors import NearestNeighbors

from exlab.interface.serializer import Serializable

from .data import *
from .space import Space
from .dataspace import DataSpace


class MultiColSpace(Space):
    def __init__(self, spaceManager, spaces, parent=Space):
        dim = sum([space.dim for space in spaces])
        parent.__init__(self, spaceManager, dim, spaces=spaces)

        # Hook when spaces are modified and this one has to be recomputed
        for space in self.spaces:
            space.childrenSpaces.append(self)
        
        rowAggregations = [space.rowAggregation for space in self.spaces]
        self.incoherentRowsAlignement = np.sum(
            rowAggregations) != 0 and np.sum(rowAggregations) < len(rowAggregations)
    
    def _sid(self, serializer):
        dict_ = serializer.serialize(
            self, ['options'], foreigns=['spaces', 'spaceManager'])
        return dict_
        
    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['spaces', 'options'], foreigns=['spaceManager'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            spaces = [serializer.deserialize(spaceData) for spaceData in dict_.get('spaces', [])]
            obj = serializer.deserialize(dict_.get('spaceManager')).multiColSpace(spaces)

        # Operations
        # for spaceData in dict_.get('spaces', []):
        #     # existing = [s for s in obj.spaces]
        #     space = serializer.deserialize(spaceData)

        return super()._deserialize(dict_, serializer, obj)

    # def _postDeserialize(self, dict_, serializer):
    #     super()._postDeserialize(dict_, serializer)

    @staticmethod
    def create(spaces, spaceManager=None):
        if not spaces:
            if spaceManager:
                return spaceManager.multiColSpace([])
            return None

        spaceManager = list(set([space.spaceManager for space in spaces]))
        if len(spaceManager) > 1:
            raise Exception(
                "All spaces should lay within the same space manager")
        s = spaceManager[0].multiColSpace(spaces)
        return s

    def icon(self):
        return super().icon() + '∥'

    def boundedProperty(self):
        return '⇉' + '👁' if self.observable() else '' + '🕹' if self.primitive() else ''

    def colStr(self):
        cols = ' '.join([space.toStr(2) for space in self.spaces])
        return f'{self.boundedProperty()}[{cols}]'

    def observable(self):
        # All sub spaces must be observable
        return sum([s.observable() for s in self.spaces]) == len(self.spaces)

    def primitive(self):
        # All sub spaces must be primitive
        return sum([s.primitive() for s in self.spaces]) == len(self.spaces) and not self.null()

    def createDataSpace(self, spaceManager=None, kind=None):
        spaces = [s.convertTo(spaceManager, kind, True) for s in self.spaces]
        return MultiColDataSpace.create(spaces, spaceManager=spaceManager)

    def createLinkedSpace(self, spaceManager=None, kind=None):
        spaces = [s.convertTo(spaceManager, kind, False) for s in self.spaces]
        return MultiColSpace.create(spaces, spaceManager=spaceManager)


class MultiColDataSpace(MultiColSpace, DataSpace):
    def __init__(self, spaceManager, spaces):
        MultiColSpace.__init__(self, spaceManager, spaces, parent=DataSpace)
    
    def _sid(self, serializer):
        return MultiColSpace._sid(self, serializer)

    def _serialize(self, serializer):
        dict_ = MultiColSpace._serialize(self, serializer)
        dict_.update(serializer.serialize(
            self, ['_spaceWeights']))
        return dict_

    # @classmethod
    # def _deserialize(cls, dict_, serializer, obj=None):
    #     return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        MultiColSpace._postDeserialize(self, dict_, serializer)
        self._spaceWeights = serializer.deserialize(dict_.get('spaceWeights', []))
        self._validate()
        self._updateSpaceWeights()

    def addPoint(self, point, idx, cost=None, action=False):
        """Add a point in the space and if valid return id."""
        # TODO
        if point.space != self:
            return -1

        for part in point.flat():
            part.space.addPoint(part, idx, cost=cost, action=action)

    def invalidate(self):
        Space.invalidate(self)
        self.data = np.array([])
        self.costs = np.array([])
        self.ids = np.array([])
        self.lids = np.array([], dtype=np.int32)
        self.actions = []
        self._number = 0

        '''self.costs = np.zeros([Space.RESERVE_STEP])
        self.data = np.zeros([Space.RESERVE_STEP, self.dim])
        self.ids = np.zeros([Space.RESERVE_STEP], dtype=np.int16)  # execution id (order)
        self.lids = np.full([Space.RESERVE_STEP], -1, dtype=np.int16)  # reverse execution order'''

    def _preValidate(self):
        if not Space._preValidate(self):
            return False
        if not self.spaces:
            return False
        return self._mergeData(col=True)

    def _mergeData(self, col=True):
        for space in self.spaces:
            space._validate()
        if col:
            ids = list(set.intersection(
                *[set(space.ids[:space._number]) for space in self.spaces]))
        else:
            ids = np.arange(np.sum([space.number for space in self.spaces]))
        if len(ids) == len(self.ids) or len(ids) == 0:
            # self.invalid = False
            return True

        # self._bounds = [x for space in self.spaces for x in space._bounds]
        #
        # if len(ids) == len(self.ids):
        #     self.invalid = False
        #     return

        if col:
            self.ids = np.array(ids)
            self.actions = list(set.intersection(
                *[set(space.actions) for space in self.spaces]))
            length = max(
                list(set.union(*[set(space.ids[:space._number]) for space in self.spaces]))) + 1

            def getData(space, length):
                n = np.zeros([length, space.dim])
                '''print(space.data.size)
                print(space.ids.size)
                print(space._number)
                print("HELLLLO")
                print(space.data[:space._number].size)
                print(n[space.ids[:space._number]].size)'''
                n[space.ids[:space._number]] = space.data[:space._number]
                return n

            def getCosts(space, length):
                n = np.zeros([length])
                n[space.ids[:space._number]] = space.costs[:space._number]
                return n

            #print("Compute {} {}".format(length, space._number))
            self.data = np.concatenate(
                [getData(space, length) for space in self.spaces], axis=1)[self.ids]
            self.costs = np.concatenate(
                [getCosts(space, length) for space in self.spaces], axis=0)[self.ids]
            self.lids = np.full([length], -1, dtype=np.int16)
            self.lids[self.ids] = np.arange(len(self.ids))
        else:
            self.ids = ids
            self.actions = np.hstack(
                [space.actions[:space.number] for space in self.spaces])
            self.data = np.vstack([space.data[:space.number]
                                   for space in self.spaces])
            self.costs = np.hstack([space.costs[:space.number]
                                    for space in self.spaces])
            self.lids = np.array(ids, dtype=np.int32)
            # lidsNZero = [np.nonzero(space.lids != -1) for space in self.spaces]
            # self.lids = np.full([np.max(lidsNZero) + 1], -1, dtype=np.int32)
            # pos = 0
            # for space, nz in zip(self.spaces, lidsNZero):
            #     self.lids[nz] = space.lids[nz] + pos
            #     pos += space.number

        self.contiguous = all(space.contiguous for space in self.spaces)

        #print(self.data)
        self._number = len(self.ids)
        return True

    def _postValidate(self):
        DataSpace._postValidate(self)
        # if random.uniform(0, 1) < 0.1:
        self._updateSpaceWeights()


class MultiRowDataSpace(DataSpace):
    def __init__(self, spaceManager, spaces):
        assert len(spaces) > 0
        if len(set([space.dim for space in spaces])) != 1:
            raise Exception("All spaces should have the same dimension")
        dim = spaces[0].dim
        super().__init__(spaceManager, dim, native=spaces[0], spaces=spaces)
        self.rowAggregation = True
        self.abstract = True

        # Hook when spaces are modified and this one has to be recomputed
        for space in self.spaces:
            space.childrenSpaces.append(self)
    
    def _sid(self, serializer):
        return MultiColSpace._sid(self, serializer)
    
    def _serialize(self, serializer):
        dict_ = MultiColSpace._serialize(self, serializer)
        dict_.update(serializer.serialize(
            self, ['_spaceWeights']))
        return dict_

    # @classmethod
    # def _deserialize(cls, dict_, serializer, obj=None):
    #     return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        MultiColSpace._postDeserialize(self, dict_, serializer)
        self._spaceWeights = serializer.deserialize(dict_.get('spaceWeights', []))
        self._validate()
        self._updateSpaceWeights()

    @staticmethod
    def create(spaces):
        if not spaces:
            return None
        spaceManager = list(set([space.spaceManager for space in spaces]))
        if len(spaceManager) > 1:
            raise Exception(
                "All spaces should lay within the same space manager")
        dataSpaces = list(set([isinstance(space, DataSpace)
                               for space in spaces]))
        if len(dataSpaces) > 1:
            raise Exception("All spaces should be a DataSpace or none")
        return spaceManager[0].multiRowSpace(spaces, data=dataSpaces[0])

    def toStr(self, short=False):
        spaces = '>'.join([space.toStr(2) for space in self.spaces])
        if short:
            return f"({spaces})({self.dim})@⛁"
        return f"@⛁({spaces})({self.dim})"
    
    def applyTo(self, entity):
        if not entity:
            return self
        return next((row for row in self.rows if row.matchesEntity(entity)), self)

    # @property
    # def colSpaces(self):
    #     return [self.spaces[0]]

    # @property
    # def colSpacesFull(self):
    #     return [self]

    # @property
    # def rowSpaces(self):
    #     return self.spaces

    @property
    def cols(self):
        return [self]

    # @property
    # def rows(self):
    #     return self.spaces + [row for row in self.allowedSimilarRows if row not in self]

    @property
    def baseCols(self):
        return [self.spaces[0]]

    @property
    def baseRows(self):
        return self.spaces

    @property
    def groupedCols(self):
        return [tuple([s for colSpace in space.groupedCols for s in colSpace[0]] for space in self.spaces)]

    def observable(self):
        # All sub spaces must be observable
        return self.spaces[0].observable()

    def primitive(self):
        # All sub spaces must be primitive
        return self.spaces[0].primitive()

    def invalidate(self):
        MultiColDataSpace.invalidate(self)

    def _preValidate(self):
        if not Space._preValidate(self):
            return False
        if len(self.spaces) == 0:
            return False
        return MultiColDataSpace._mergeData(self, col=False)

    def _postValidate(self):
        DataSpace._postValidate(self)
        # if random.uniform(0, 1) < 0.1:
        self._updateSpaceWeights()



import copy
import random
import numpy as np

# from ...utils.serializer import getReference
# from ...data.dataset import Action
from dinos.data.data import Action, SingleAction, ActionList
# from dinos.data.space import *
from dinos.utils.move import MoveConfig

from .strategy import Strategy


class RandomStrategy(Strategy):
    """The random exploration of actions strategy."""

    def __init__(self, agent, name=None, performer=None, planner=None, options={}):
        super().__init__(agent, name=name, planner=planner, performer=performer,
                         options=options)
        self.exploreNonPrimitive = self.options.get(
            'exploreNonPrimitive', True)

    def _serialize(self, serializer):
        dict_ = super()._serialize(serializer)
        dict_.update(serializer.serialize(self))
        return dict_

    # @classmethod
    # def _deserialize(cls, dict_, agent, options=None, obj=None):
    #     obj = obj if obj else cls(agent, dict_.get(
    #         'name'), options=dict_.get('options', {}))
    #     obj = Strategy._deserialize(dict_, agent, options=options, obj=obj)
    #     return obj

    def _runIteration(self, config):
        # Test a random action
        self.testRandomAction(config)

    # test modifications on both simple and complex agents
    def testRandomAction(self, config=MoveConfig(), actionSpaces=None, zero=False):
        """Build and test a random action."""
        if not actionSpaces:
            actionSpaces = self.agent.explorableSpaces(onlyPrimitives=not self.exploreNonPrimitive)
        if self.agent.dataset:
            actionSpaces = self.agent.dataset.controllableSpaces(actionSpaces)
        space = random.choice(actionSpaces)

        # self.logger.debug("Random action performed in {} selected among {} spaces"
        #                   .format(getReference(space), len(actionSpaces)), 'BABLG')
        '''#space = self.env.dataset.actionExplorationSpaces[0]
        n = 0
        actions = []
        # if self.complex_actions <= 0:
        cost = 1.0
        r = random.random()
        # Keep chaining current actions with primitives to build a complex action
        while len(actions) == 0 or r < cost:
            action = SingleAction(space, [random.uniform(minb, maxb) for minb, maxb in space.bounds])
            actions.append(action)
            n += 1
            r = random.random()
            # dev
            cost = 0.
            #cost = 1.0 / (self.k ** n)
        # else:
        #     n = self.complex_actions
        #     for i in range(self.complex_actions):
        #         action = SingleAction(space, [random.uniform(minb, maxb) for minb, maxb in space.bounds])
        #         actions.append(action)'''

        actions = []
        if zero:
            action = space.zero()
        else:
            action = SingleAction(space, [random.uniform(minb, maxb) for minb, maxb in space.bounds])
        actions.append(action.setRelative(True))

        actions = ActionList(*actions)
        config.result.action = action
        # print(f'DOING {action} {action.relative}')
        self.testActions(actions, config)



from dinos.utils.move import MoveConfig


class Challenge(object):
    def __init__(self, method, name=None):
        self.method = method
        self.name = name if name else method.__name__
        self.scene = None

    def __repr__(self):
        return f'Challenge {self.name}'

    def attempt(self, agent, video=False):
        self.agent = agent
        if video:
            self.world.record()
        self.method(self)
        if video:
            self.world.record(False)
            return self.world.video()

    @property
    def world(self):
        return self.scene.world

    def reach(self, goal, config=MoveConfig()):
        origin = self.world.observe(spaces=goal.space.flatSpaces)
        self.agent.reachGoal(goal, config)
        final = self.world.observe(spaces=goal.space.flatSpaces)
        print(f'{self} result: reached {final}, asked {goal} (from {origin})')


class SceneSetup(object):
    CLASS_ENDNAME = 'Scene'
    environmentClass = None

    def __init__(self, environment):
        self.environment = environment

        self.challenges = []
        self.tests = []
        self.testIds = {}
        self._configure()
    
    @classmethod
    def init(cls):
        return cls.environmentClass(cls)
    
    @property
    def world(self):
        return self.environment.world

    def serialize(self, options={}):
        dict_ = {'id': self.__class__.__name__}
        return dict_
    
    @property
    def name(self):
        name = self.__class__.__name__
        if name.endswith(self.CLASS_ENDNAME):
            name = name[:-len(self.CLASS_ENDNAME)]
        return name

    def __repr__(self):
        return f'SceneSetup {self.__class__.__name__} for env {self.environment.envname}'

    def _configure(self):
        pass

    def _setup(self):
        pass

    def setup(self):
        self._setup()

    def _setupTests(self):
        pass

    def setupTests(self):
        self._setupTests()

    # Before iteration / episode
    def setupIteration(self, config=MoveConfig()):
        pass

    def setupEpisode(self, config=MoveConfig()):
        pass

    def setupPreTest(self, test=None):
        pass

    def reset(self):
        self._reset()

    def _reset(self):
        pass

    def _draw(self, base, drawOptions):
        pass

    def _preIteration(self):
        pass

    def reward(self, action):
        return 0.

    def addChallenge(self, challenge):
        if challenge not in self.challenges:
            self.challenges.append(challenge)
            challenge.scene = self

    def addTest(self, test):
        if test not in self.tests:
            self.tests.append(test)
            test.scene = self
            test._bindId()



import copy
from enum import Enum
from contextlib import contextmanager

from .space import SpaceKind


class AMTElementKind(Enum):
    NONE = 'none'
    ENTITY = 'entity'
    PROPERTY = 'property'


class AMTBase(object):
    NAME = 'Base'
    KIND = AMTElementKind.NONE

    def __init__(self, context, parent=None, element=None):
        self.context = context
        self.element = element
        # self.parents = set()
        self.children = set()
        # self.addParent(parent)

    @property
    def abstract(self):
        return True

    def copy(self, context=None):
        obj = self.__class__.__new__(self.__class__)
        obj.__dict__.update(self.__dict__)
        # if not context:
        #     if self.context == self:
        #         context = obj
        #     else:
        #         context = self.context.copy()
        if obj.element:
            obj.element = copy.copy(obj.element)
        obj.children = [child.copy(context) for child in obj.children]
        return obj

    def addChild(self, child):
        if child not in self.children:
            self.children.add(child)
            child.parent = self

    def addChildren(self, children):
        for child in children:
            self.addChild(child)

    # def addParent(self, parent):
    #     if not parent:
    #         return
    #     if parent not in self.parents:
    #         self.parents.add(parent)
    #         parent.addChild(self)

    # Visitor
    def abstractedElements(self, kind=None):
        elements = set()
        if self.abstract and self.element and (not kind or self.KIND.value == kind.value):
            elements.add(self.element)
        elements |= set(
            e for child in self.children for e in child.abstractedElements())
        return elements

    def assignableElements(self):
        assignables = {}
        for child in self.children:
            assignables.update(child.assignableElements())
        if self.abstract and self.element:
            assignables[self.element] = set(self.element.assignables)
        return assignables
    
    def abstractedEntities(self):
        return self.abstractedElements(kind=AMTElementKind.ENTITY)

    def abstractedEntityProperties(self):
        elements = {}
        if self.element and self.KIND.value == AMTElementKind.PROPERTY.value and list(self.children)[0].abstract:
            entity = list(self.children)[0].element
            elements[entity] = elements.get(entity, set()) | set([self.element])
        for child in self.children:
            for entity, childElements in child.abstractedEntityProperties().items():
                elements[entity] = elements.get(entity, set()) | childElements
        return elements

    # def assign(self, element, value):
    #     tree = self.copy()
    #     tree.assignInplace(element, value)
    #     return tree

    def assignInplace(self, element, value):
        if self.element == element:
            self.element = AMTElement(
                self.context, self.__class__, assigned=value)
        for child in self.children:
            child.assignInplace(element, value)

    def get(self, spaceManager):
        pass

    # Representation
    def __repr__(self):
        return self.toStr()

    def toStr(self):
        return f"{', '.join([item.toStr() for item in self.children])}"

    def name(self):
        if self.element:
            return self.element.name()
        raise Exception('Not a namable abstract object!')

    def fullname(self):
        return self._fullnamePattern(self.name())

    @staticmethod
    def _namePattern(obj, index):
        raise Exception('Not a namable abstract object!')

    def _fullnamePattern(self, name):
        raise Exception('Not a namable abstract object!')


class AMT(AMTBase):
    def __init__(self, context=None):
        super().__init__(context if context else self)

        self.fullnames = {-1: '__base__'}
        self.names = {-1: '__base__'}
        self.matching = {}
        self.sameRowsSpaces = True

        if context:
            context.addChild(self)

    def expr(self, item):
        obj = AMTExpr(self.context, item)
        self.addChild(obj)
        return obj

    def entity(self, assigned=None, element=None):
        obj = AMTEntity(self.context, assigned, element=element)
        return obj

    def multicol(self, *args):
        obj = AMTMultiCol(self.context, *args)
        self.addChild(obj)
        return obj

    def get(self, spaceManager):
        assert len(self.children) == 1

        # if self.context.sameRowsSpaces and len(spaces) == 1:
        #     spaces = [spaces[0]] * len(list(allAssignables.values())[0])

        return list(self.children)[0].get(spaceManager)

    # Abstract
    def abstractEntity(self, entity, abstractNewElements=True):
        if id(entity) in self.context.matching:
            return self.entity(element=self.context.matching[id(entity)])
        if abstractNewElements:
            obj = self.entity()
        else:
            obj = self.entity(entity)
        self.context.matching[id(entity)] = obj.element
        obj.element.assignable(entity)
        return obj

    def abstractProperty(self, property_, kind=SpaceKind.BASIC, abstractNewElements=True, appendToChildren=False):
        entity = self.abstractEntity(
            property_.entity, abstractNewElements=abstractNewElements)
        if id(property_) in self.context.matching:
            return entity.propertyItem(element=self.context.matching[id(property_)], kind=kind)
        obj = entity.propertyItem(propertyName=property_.name, kind=kind)
        self.context.matching[id(property_)] = obj.element
        if appendToChildren:
            self.addChild(obj)
        return obj

    def abstractSpace(self, space, abstractNewElements=True, appendToChildren=False):
        if not space.aggregation:
            if space.spaces[0] == space:
                property = space.boundProperty
                kind = space.kind
                # entity = self.abstractEntity(property.entity, abstractNewElements=abstractNewElements)
                expr = self.abstractProperty(
                    property, kind=kind, abstractNewElements=abstractNewElements)
            else:
                pass
        else:
            expr = self.multicol(
                *[self.abstractSpace(s, abstractNewElements=abstractNewElements) for s in space.flatColsWithoutMultiRows])
        if appendToChildren:
            self.addChild(expr)
        return expr


class AMTExpr(AMTBase):
    def __init__(self, context, item):
        super().__init__(context)
        self.addChild(item)


class AMTMultiCol(AMTBase):
    def __init__(self, context, *args):
        super().__init__(context)
        self.addChildren(args)

    def get(self, spaceManager):
        spaces = []
        for child in self.children:
            s = child.get(spaceManager)
            s.append(s)
        space = spaceManager.multiColSpace(spaces)
        return space

    def toStr(self):
        return f"[{'+'.join([item.toStr() for item in self.children])}]"


class AMTElementLinked(AMTBase):
    def __init__(self, context, assigned=None, parent=None, element=None):
        if not element:
            element = AMTElement(context, self.__class__, assigned=assigned)
        super().__init__(context, parent=parent, element=element)

    @property
    def abstract(self):
        return self.element.abstract

    def assignable(self, obj):
        self.element.assignable(obj)

    def unassignable(self, obj):
        self.element.unassignable(obj)

    def toStr(self):
        return self.fullname()


class AMTEntity(AMTElementLinked):
    NAME = 'Entity'
    KIND = AMTElementKind.ENTITY

    def __init__(self, context, entity=None, element=None):
        super().__init__(context, element=element, assigned=entity)

    def propertyItem(self, propertyName=None, element=None, kind=SpaceKind.BASIC):
        return AMTProperty(self.context, self, propertyName=propertyName, element=element, kind=kind)

    def get(self, spaceManager):
        return self.element.assignables

    @staticmethod
    def _namePattern(obj, index):
        return f"({chr(ord('A') + index)})"

    def _fullnamePattern(self, name):
        return name


class AMTProperty(AMTElementLinked):
    NAME = 'Property'
    KIND = AMTElementKind.PROPERTY

    def __init__(self, context, entity, propertyName=None, element=None, kind=SpaceKind.BASIC):
        super().__init__(context, element=element, assigned=propertyName)
        self.kind = kind
        self.addChild(entity)

    def toStr(self):
        return self.fullname()

    @property
    def entity(self):
        return list(self.children)[0]

    def get(self, spaceManager):
        if len(self.element.assignables) == 0:
            raise Exception(
                f'{self.element} has no possible assignable values!')

        if self.context.sameRowsSpaces:
            allAssignables = self.context.assignableElements()
            rows = 1
            if len(allAssignables) > 1:
                raise Exception(
                    'Only one assignable may vary with \'sameRowsSpaces\' enabled')
            if len(allAssignables) == 1:
                rows = len(list(allAssignables.values())[0])

        properties = []
        for propertyName in list(self.element.assignables):
            for entity in self.entity.get(spaceManager):
                prop = entity.propertyItem(propertyName)
                if prop is None:
                    raise Exception(
                        f'{entity} has no property {propertyName}!')
                properties.append(prop)
        spaces = [spaceManager.spaceSearch(
            property=property, kind=self.kind) for property in properties]

        if self.context.sameRowsSpaces and len(spaces) == 1:
            spaces = [spaces[0]] * rows

        return spaceManager.multiRowSpace(spaces)

    @staticmethod
    def _namePattern(obj, index):
        n = chr(ord('a') + index)
        return f'({n})'

    def _fullnamePattern(self, name):
        return f'{self.entity.name()}.{name}'


class AMTElement(object):
    def __init__(self, context, cls, assigned=None):
        # if assigned:
        #     name = repr(assigned)
        self.context = context
        self.cls = cls
        self._assignables = set()
        self.unassignables = set()
        self.assigned = assigned
        if self.assigned:
            self.assignable(self.assigned)

    def name(self):
        if id(self) in self.context.names:
            return self.context.names[id(self)]
        if self.assigned:
            name = str(self.assigned)
            self.context.names[id(self)] = name
            # self.context.fullnames[id(self)] = self.cls._fullnamePattern(self, self._name)
            return name

        name = '__base__'
        i = 0
        while name in self.context.names.values():
            name = self.cls._namePattern(self, i)
            # fullname = self.cls._fullnamePattern(self, name)
            i += 1

        self.context.names[id(self)] = name
        # self.context.fullnames[id(self)] = fullname
        return name

    @property
    def abstract(self):
        return self.assigned is None

    def assignable(self, obj):
        self._assignables.add(obj)
        if obj in self.unassignables:
            self.unassignables.remove(obj)
        if len(self._assignables) > 1:
            self.assigned = None

    def unassignable(self, obj):
        self.unassignables.add(obj)
        if obj in self._assignables:
            self._assignables.remove(obj)

    @property
    def assignables(self):
        if self.assigned:
            return [self.assigned]
        return self._assignables

    @contextmanager
    def assign(self, assigned):
        previously = self.assigned
        try:
            self.assigned = assigned
            yield self
        finally:
            self.assigned = previously

    def __repr__(self):
        return f'{self.cls.NAME} {self.name()}'



from .internal_engine import InternalEngine, GraphicalEngine, DrawOptions

import numpy as np

from PIL import Image
from scipy.spatial.distance import euclidean

# PyMunk
import pymunk
import pymunk.pygame_util
from pymunk import Vec2d

# PyGame
import pygame
import pygame.locals as pglocals
from pygame.color import THECOLORS

# Panda 3D
from panda3d.core import Point3, Vec3, ClockObject, loadPrcFileData, AntialiasAttrib, PointLight, AmbientLight, DirectionalLight, Spotlight, Filename, getModelPath


class PymunkEngine(InternalEngine):
    ENGINE = 'Pymunk'

    # def __init__(self, environment, options={}):
    #     super().__init__(environment, options)

    def setupGraphicalEngines(self):
        self.registerGraphicalEngine(PyGameGE, True)
        self.registerGraphicalEngine(Panda3DGE)

    def initPhysics(self):
        # Physics stuff
        self.physics = pymunk.Space()
        self.physics.gravity = (0.0, 0.0)
        self.physics.damping = 0.1

    
    # def reset(self):
    #     for obj in self.agents + self.objs:
    #         if isinstance(obj, Physical):
    #             obj.body.position = obj.coords_init
    #             obj.body.angle = 0
    #             obj.body.angular_velocity = 0
    #             obj.body.velocity = Vec2d(0.0, 0.0)
    #     self.setupIteration()

    def _updatePhysics(self, dt):
        self.physics.step(dt)

    def checkSegmentOccupancy(self, x1, y1, x2, y2, width=0, exceptions=[]):
        exceptionShapes = [entity.shape for entity in exceptions]
        pos = Vec2d(x1, y1)
        dest = Vec2d(x2, y2)
        hits = self.physics.segment_query(
            pos, dest, width, pymunk.ShapeFilter())
        filteredHits = [
            info for info in hits if info.shape not in exceptionShapes]
        return len(filteredHits) > 0

    def checkCaseOccupancy(self, x, y, w, h, exceptions=[]):
        exceptionShapes = [entity.shape for entity in exceptions]

        body = pymunk.Body(1., 1.)
        body.position = x + w / 2, y + h / 2
        shape = pymunk.Poly(
            body, [(-w/2, -h/2), (-w/2, h/2), (w/2, h/2), (w/2, -h/2), (-w/2, -h/2)])

        hits = self.physics.shape_query(shape)
        filteredHits = [
            info for info in hits if info.shape not in exceptionShapes]
        return len(filteredHits) > 0

    def checkGridOccupancy(self, x1, y1, x2, y2, caseW, caseH, exceptions=[]):
        numX = (x2 - x1) // caseW
        numY = (y2 - y1) // caseH
        grid = np.zeros((numY, numX))

        for i in range(numY):
            for j in range(numX):
                grid[i, j] = self.checkCaseOccupancy(
                    x1 + caseW*j, y1 + caseH*i, caseW, caseH, exceptions=exceptions)

        return grid

    def customStep(self, duration=0.):
        if self.displayThisFrame:
            for g in self.graphicalEngines:
                g.customStep(duration)


class PyGameGE(GraphicalEngine):
    def _initGraphics(self):
        pygame.init()
        self._updateGraphics(False)
        self.clock = pygame.time.Clock()
        self.draw_options = pymunk.pygame_util.DrawOptions(self.screen)
        self.font = pygame.font.SysFont("monospace", 13)

    def _updateGraphics(self, oldgui):
        if self.gui:
            self.screen = pygame.display.set_mode(self.DISPLAY_SIZE)
        else:
            if oldgui:
                pygame.display.quit()
            self.screen = pygame.Surface(self.DISPLAY_SIZE)

    def _image(self):
        self.engine.updatePhysics()
        self.drawScreen()

        pil_string_image = pygame.image.tostring(self.screen, "RGBA", False)
        pil_image = Image.frombytes("RGBA", self.DISPLAY_SIZE, pil_string_image)

        return pil_image

    def flipScreen(self, delay=False):
        if self.gui:
            pygame.display.flip()
            pygame.event.get()
            if delay:
                self.clock.tick(50*(self.engine.speedupFrames if self.engine.speedupFrames is not None else 1000000))
                pygame.display.set_caption(f"{self.engine.windowTitle()} | GUI fps: {self.clock.get_fps():.0f}")

    def _hide(self):
        pygame.display.set_caption(
            f"{self.engine.windowTitle()} | GUI Stopped (simulation may continue in background)")

    def _show(self):
        pygame.display.set_caption(
            f"{self.engine.windowTitle()} | GUI Ready (simulation on pause)")

    def _drawScreen(self, duration=0.):
        self.screen.fill(THECOLORS["white"])  # Clear screen
        #self.physics.debug_draw(self.draw_options) 
        drawOptions = PMDrawOptions(self, False)
        self._drawSelfScene(self.screen, drawOptions)
        for entity in self.environment.world.cascadingChildren():
            entity.draw(self.screen, drawOptions)
        if duration:
            label_acting = self.font.render("Acting", 1, (255, 0, 0))
            self.screen.blit(label_acting, (24, 24))
        else:
            pass
            # label_mode = self.font.render("Current outcome space: / [J] to change", 1,
            #                               (0, 0, 0))
            # self.screen.blit(label_mode, (24, 24))
    
    def customStep(self, duration=0.):
        if self.gui and duration is None:
            for event in pygame.event.get():
                if event.type == pglocals.QUIT:
                    self.running = False
                elif event.type == pglocals.KEYDOWN and event.key == pglocals.K_ESCAPE:
                    self.running = False
                elif event.type == pglocals.KEYDOWN and event.key == pglocals.K_j:
                    # self.agents[0].execute_action((1.0, 0.0))
                    # change mode
                    if self.current_space == self.env.dataset.outcomeSpaces[-1]:
                        self.current_space = self.env.dataset.outcomeSpaces[0]
                    else:
                        self.current_space = self.env.dataset.outcomeSpaces[self.env.dataset.outcomeSpaces.index(
                            self.current_space) + 1]
                elif event.type == pygame.MOUSEBUTTONUP:
                    # convert to local agent coordinates
                    pos = self.propertyItem(self.current_space.property).convert_to_feature(
                        pymunk.pygame_util.from_pygame(event.pos, self.screen))
                    self.experiment.exploit(
                        pos, self.env.dataset.outcomeSpaces.index(self.current_space))

        # Draw stuff
        if self.gui or self._record:
            self.drawScreen(duration)


class Panda3DGE(GraphicalEngine):
    def pandaResourcesPath(self, filepath=''):
        return Filename.fromOsSpecific(str(self.engine.environment.resourcesPath(filepath).absolute())).getFullpath()

    def _initGraphics(self):
        self._updateGraphics(False)

        from direct.showbase import ShowBase
        self.base = ShowBase.ShowBase()

        getModelPath().appendDirectory(self.pandaResourcesPath())
        print(getModelPath())

        # if self.gui:
        self.base.cam.setPos(50, 50, 50)
        self.base.cam.lookAt(0, 0, 0)
        # self.base.setFrameRateMeter(True)

        # self.base.render.setAntialias(AntialiasAttrib.MMultisample)
        # self.base.render.setAntialias(AntialiasAttrib.MMultisample, 4)
        self.base.render.setAntialias(AntialiasAttrib.MAuto)

        # teapot = loader.loadModel('teapot')
        # teapot.reparentTo(self.base.render)
        # teapot.setPos(100, 100, 50)
        # teapot.setScale(20.)
        # teapotMovement = teapot.hprInterval(50, Point3(0, 360, 360))
        # teapotMovement.loop()

        self.sun = self.base.render.attachNewNode(Spotlight("Spot"))
        self.sun.node().setScene(self.base.render)
        self.sun.node().setShadowCaster(True, 4096, 4096)
        # self.sun.node().showFrustum()
        self.sun.node().getLens().setFov(60)
        self.sun.node().getLens().setNearFar(10, 2000)
        self.sun.node().setColor((1., 1., 1., 1))
        self.base.render.setLight(self.sun)

        # self.sun = DirectionalLight('sun')
        # self.sun.setColor((0.8, 0.8, 0.5, 1))
        # self.sun = PointLight('sun')
        # self.sun.setColor((0.9, 0.9, 0.9, 1))
        # self.sun.attenuation = (1, 0, 0)
        # self.sun.setShadowCaster(True, 64, 64)

        # self.sunp = self.base.render.attachNewNode(self.sun)
        # self.sunp.setHpr(50, -60, 0)
        self.sun.setPos(500, 550, 1200)
        self.sun.lookAt(300, 300, 0)

        alight = AmbientLight('alight')
        alight.setColor((0.3, 0.3, 0.3, 1))
        alnp = self.base.render.attachNewNode(alight)
        self.base.render.setLight(alnp)

        self.base.render.setShaderAuto()

    def _updateGraphics(self, oldgui):
        self.gui = True
        if self.gui:
            loadPrcFileData("", "window-title Hello")
            loadPrcFileData("", "win-size 1800 600")
            loadPrcFileData("", "sync-video 0")
        else:
            if oldgui:
                pass
            loadPrcFileData("", "window-type offscreen")
        loadPrcFileData("", "framebuffer-multisample 1")
        loadPrcFileData("", "multisamples 2")

    def _image(self):
        self.engine.updatePhysics()
        self.drawScreen()

        self.base.graphicsEngine.renderFrame()
        self.base.graphicsEngine.renderFrame()

        FILENAME = 'tmp.png'
        self.base.screenshot(FILENAME, defaultFilename=False)
        imageWithAlpha = Image.open(FILENAME).convert('RGBA')

        image = Image.new("RGB", imageWithAlpha.size, (128, 128, 128))
        image.paste(imageWithAlpha, mask=imageWithAlpha.split()[3])

        return image

    def flipScreen(self, delay=False):
        if self.gui:
            self.base.graphicsEngine.renderFrame()
            if delay:
                pass

    def _hide(self):
        pass

    def _show(self):
        pass

    def _drawScreen(self, duration=0.):
        drawOptions = PMDrawOptions(self, True)
        self._drawSelfScene(self.base, drawOptions)
        for entity in self.environment.world.cascadingChildren():
            entity.draw(self.base, drawOptions)

    def customStep(self, duration=0.):
        # if self.gui and duration is None:
        #     for event in pygame.event.get():
        #         if event.type == pglocals.QUIT:
        #             self.running = False
        #         elif event.type == pglocals.KEYDOWN and event.key == pglocals.K_ESCAPE:
        #             self.running = False
        #         elif event.type == pglocals.KEYDOWN and event.key == pglocals.K_j:
        #             # self.agents[0].execute_action((1.0, 0.0))
        #             # change mode
        #             if self.current_space == self.env.dataset.outcomeSpaces[-1]:
        #                 self.current_space = self.env.dataset.outcomeSpaces[0]
        #             else:
        #                 self.current_space = self.env.dataset.outcomeSpaces[self.env.dataset.outcomeSpaces.index(
        #                     self.current_space) + 1]
        #         elif event.type == pygame.MOUSEBUTTONUP:
        #             # convert to local agent coordinates
        #             pos = self.propertyItem(self.current_space.property).convert_to_feature(
        #                 pymunk.pygame_util.from_pygame(event.pos, self.screen))
        #             self.experiment.exploit(
        #                 pos, self.env.dataset.outcomeSpaces.index(self.current_space))

        # Draw stuff
        if self.gui or self._record:
            self.drawScreen(duration)


class PMDrawOptions(DrawOptions):
    def __init__(self, graphicalEngine, panda):
        super().__init__(graphicalEngine)
        self.panda = panda
        self.pygame = not panda



'''
    File name: loaders.py
    Author: Nicolas Duminy, Alexandre Manoury
    Python Version: 3.6
'''

import scipy.optimize
import numpy as np
import subprocess
import warnings
import logging
import random
import pickle
import json
import time
import csv
import sys
import os

from scipy.optimize import OptimizeResult
from sklearn.linear_model import LinearRegression, RANSACRegressor
from sklearn import linear_model
from numpy.linalg import LinAlgError
import matplotlib.pyplot as plt


from dinos.data import operations


def uniformRowSampling(array, probs):
    """
    Choses a row in an array using a weighted random choice
    Exemple:
        uniformRowSampling(['a', 'b'], [2, 1])
        'a' will have twice the chances to appear than 'b'
    """
    probs /= np.sum(probs)
    return array[np.random.choice(len(array), 1, p=probs)[0]]


def uniformSampling(probs):
    """Uniform sampler."""
    probs /= np.sum(probs)
    return np.random.choice(len(probs), 1, p=probs)[0]


def linearValue(start, end, position):
    return start + (end - start) * max(min(position, 1), 0)


def first(list_, default=None):
    if list_:
        return list_[0]
    return default


def iterrange(collection, range_=(-1, -1), key=lambda item: item[0]):
    if isinstance(collection, dict):
        collection = collection.items()
    return list(filter(lambda item: key(item) > range_[0] and (key(item) <= range_[1] or range_[1] == -1), collection))


def popn(list_, number=1, fromEnd=False):
    """
    pops n elements from a list
    """
    ret = []
    for _ in range(number):
        if fromEnd:
            ret.append(list_.pop())
        else:
            ret.append(list_.pop(0))
    return ret


def sigmoid(x):
    """Compute a sigmoid."""
    y = 1.0 / (1.0 + np.exp(-x))
    return y


def threshold(value, ratio):
    # return (1 - value * (0.5 - ratio) * 2)
    return (ratio >= 0.5) * (1 - (1 - value) * (0.5 - ratio) * 2) + (ratio < 0.5) * (1 - value * (0.5 - ratio) * 2)


def mixedSort(values1, values2, min1=None, max1=None, min2=None, max2=None):
    if not min1:
        min1 = np.min(values1)
    if not max1:
        max1 = np.max(values1)
    if not min2:
        min2 = np.min(values2)
    if not max2:
        max2 = np.max(values2)
    if min1 < max1:
        values1 = (values1 - min1) / (max1 - min1)
    if min2 < max2:
        values2 = (values2 - min2) / (max2 - min2)

    values = np.array(values1 + values2)
    indices = values.argsort()

    return indices, values


def multivariateRegression(x, y, x0, columns=None):
    """Compute a multivariate linear regression model y = f(x) using (X,y) and use it to compute f(x0)."""
    if columns is not None:
        x = x[:, columns]
        x0 = np.array(x0)[columns]
    try:
        return operations.multivariateRegression(x, y, x0)
    except ValueError as e:
        logging.critical(f"Regression failed: y is {y.shape[0]}x{y.shape[1]}d, X is {x.shape[0]}x{x.shape[1]}d and goal is {len(x0)}d ({e})")
        # return operations.multivariateRegression(x, y, x0)
    except LinAlgError as e:
        return operations.multivariateRegression(x[:1 + len(x) // 2], y[:1 + len(x) // 2], x0)


def multivariateRegressionError(x, y, x0, testSetX=None, testSetY=None, columns=None, weights=None):
    if columns is not None:
        x = x[:, columns]
        x0 = np.array(x0)[columns]
    if weights is not None:
        if columns is not None:
            weights = weights[columns]
        x = x * weights
        x0 = x0 * weights
    # if len(x.shape) == 2:
    #     plt.figure()
    #     plt.scatter(x[:, 0], x[:, 1])
    #     plt.scatter(y[:, 0], y[:, 1])
    try:
        return operations.multivariateRegressionError(x, y, x0)
    except ValueError as e:
        logging.critical(f"Regression failed: y is {y.shape[0]}x{y.shape[1]}d, X is {x.shape[0]}x{x.shape[1]}d and goal is {len(x0)}d ({e})")
        return operations.multivariateRegressionError(x[:1 + len(x) // 2], y[:1 + len(x) // 2], x0)
    except LinAlgError as e:
        # print('ERRR====')
        # print(x)
        # print(y)
        # print(x0)
        return operations.multivariateRegressionError(x[:1 + len(x) // 2], y[:1 + len(x) // 2], x0)

def mre(x, y, x0):
    # code from scipy LinearRegression:
    number = int(len(x) * 0.9)
    x_offset = np.average(x[:number], axis=0)
    y_offset = np.average(y[:number], axis=0)

    coef_ = np.linalg.lstsq(x[:number] - x_offset, y[:number] - y_offset, rcond=None)[0]
    if y.ndim == 1:
        coef_ = np.ravel(coef_)

    intercept_ = y_offset - np.dot(x_offset, coef_)
    y0 = np.array(x0).dot(coef_) + intercept_
    yp = np.sum(np.square(x.dot(coef_) + intercept_ - y), axis=1)
    error = np.sum(yp)# / np.sum(0.01 + np.square(y - np.mean(y)))
    # print('===')
    # print(x)
    # print(y)
    # print(yp)
    # print(error)
    # + 0.01 to avoid dividing by zero when all y are equal

    return y0, error, yp


def normalEquation(X, y):
    """Compute the multivariate linear regression parameters using normal equation method."""
    theta = np.zeros((X.shape[1], y.shape[1]))
    X2 = X.transpose()
    Xinv = np.linalg.pinv(X2.dot(X))
    theta = Xinv.dot(X2).dot(y)
    return theta



'''
    File name: environment.py
    Author: Alexandre Manoury
    Python Version: 3.6
'''

from PIL import Image


class Engine(object):
    PHYSICAL = False
    FREQ_SIM = 60.0
    PHYSICS_STEPS = 5

    def __init__(self, environment, options={}):
        self.environment = environment

        # Time variables
        self.iteration = 0
        self.time = 0.

        # Options
        self.options = options

        # Simu speed
        self.speedupFrames = options.get('speedup-frames', 20.)
        self.skippedFrames = options.get('skipped-frame', 50)
        self.jumpCurrentFrame = 0
        self.displayThisFrame = True

        self.graphicalEngine = None
        self.graphicalEngines = []
        self.setupGraphicalEngines()

        self.initPhysics()
    
    def setupGraphicalEngines(self):
        pass
    
    def registerGraphicalEngine(self, cls, default=False):
        g = cls(self, self.options)
        if self.graphicalEngine is None or default:
            self.graphicalEngine = g
        if g not in self.graphicalEngines:
            self.graphicalEngines.append(g)
    
    @property
    def scene(self):
        return self.environment.scene
    
    @property
    def world(self):
        return self.environment.world
    
    def windowTitle(self):
        return f'Dinos {self.environment.name} - {self.scene.name} - {self.environment.iteration}'
    
    # Graphics
    def show(self):
        self.graphicalEngine.show()

    def hide(self):
        self.graphicalEngine.hide()

    # Image / Video
    def display(self):
        return self.graphicalEngine.display()

    def render(self):
        return self.display()

    def recordOnce(self):
        return self.graphicalEngine.recordOnce()

    def record(self, record=True):
        return self.graphicalEngine.record(record)
    
    def drawScreen(self, duration=0.):
        self.graphicalEngine.drawScreen(duration)

    # Physics
    def initPhysics(self):
        pass

    def updatePhysics(self, dt):
        pass

    # Frames
    def checkJumpFrame(self):
        self.displayThisFrame = True
        if self.skippedFrames == 0:
            return True
        self.jumpCurrentFrame += 1
        if self.jumpCurrentFrame > self.skippedFrames:
            self.jumpCurrentFrame = 0
            return True
        self.displayThisFrame = False
        return False

    def simu(self):
        graphical = (
            'displayed' if self.graphicalEngine.gui else 'hidden') if self.graphicalEngine.graphical else 'graphical disabled'
        speed = f'x{self.speedupFrames}' if self.speedupFrames is not None else 'max'
        jump = 'displays all frames'
        if self.skippedFrames > 0:
            jump = f'displays 1 frame every {self.skippedFrames} frames'
        return f'Environment {self.__class__.__name__}\nGUI: {graphical} - fps {speed} - {jump}\n'

    # Main loop
    def run(self, duration=None):
        elapsedTime = 0
        self.running = True
        if not duration:
            duration = self.environment.timestep
        while self.running:
            self.checkJumpFrame()
            self.customStep(duration)

            dt = 1.0 / self.FREQ_SIM

            # Update physics
            if self.PHYSICAL:
                subdt = dt / float(self.PHYSICS_STEPS)
                for _ in range(self.PHYSICS_STEPS):
                    self.updatePhysics(subdt)

                    # Update
                    self.environment.world.update(subdt)

            # Flip screen
            for g in self.graphicalEngines:
                if self.displayThisFrame:
                    g.flipScreen(delay=True)
                g.checkRecording(self.time)
            elapsedTime += dt
            if duration and elapsedTime >= duration:
                self.running = False
        self.running = True
        for g in self.graphicalEngines:
            g.checkRecordingOnce()
    
    def customStep(self, duration=0.):
        pass


class GraphicalEngine(object):
    DISPLAY_SIZE = (600, 600)
    IMAGES_MAX = 400
    IMAGE_WIDTH = 300
    VIDEO_FPS = 3
    VIDEO_PERIOD = 1. / float(VIDEO_FPS)

    def __init__(self, engine, options):
        self.engine = engine

        # Image / Video
        self._record = False
        self._recordOnce = False
        self._images = []

        # GUI
        self._gui = False
        self.graphical = False
        self.graphicsLoaded = False
        self.gui = options.get("gui", False)
    
    def __repr__(self):
        return f'GraphicalEngine {self.__class__.__name__}'
    
    @property
    def environment(self):
        return self.engine.environment
    
    def show(self):
        self.displayGui(True)

    def hide(self):
        self.displayGui(False)
    
    def display(self):
        return self.image()

    def displayGui(self, gui=True):
        if self.gui == gui and self.graphical:
            return

        oldgui = self.gui
        self.gui = gui
        if self.gui:
            self.useGraphics(True)
            self._updateGraphics(oldgui)
            self._show()
            self.display()
        else:
            self._updateGraphics(oldgui)
            self._hide()

    def useGraphics(self, graphical=True, gui=None):
        if self.graphical != graphical:
            self.graphical = graphical
            if self.graphical and not self.graphicsLoaded:
                self.graphicsLoaded = True
                self._initGraphics()

        if gui is not None:
            self.displayGui(gui)

    def _initGraphics(self):
        pass

    def _updateGraphics(self, oldgui):
        pass

    def _hide(self):
        pass

    def _show(self):
        pass

    def _image(self):
        pass

    def _drawScreen(self, duration=0.):
        pass

    def _drawSelfScene(self, base, drawOptions):
        self._draw(base, drawOptions)
        self.environment.scene._draw(base, drawOptions)

    def _draw(self, base, drawOptions):
        pass

    def drawScreen(self, duration=0.):
        self._drawScreen(duration)

    def flipScreen(self):
        pass

    def customStep(self, duration=0.):
        pass

    def image(self):
        self.useGraphics()
        img = self._image()
        self.flipScreen()
        hsize = int((float(img.size[1])*float(self.IMAGE_WIDTH / img.size[0])))
        img = img.resize((self.IMAGE_WIDTH, hsize), Image.ANTIALIAS)
        return img

    def video(self, format='MP4V', fps=10):
        image_array = [utils.video.image_PIL2np(
            image) for image in self._images]
        utils.video.display_videojs(image_array[::])

    def _toVideo(self):
        if len(self._images) > self.IMAGES_MAX:
            self._images.pop(0)
        self._images.append(self.image())
    
    def checkRecording(self, time):
        if self._record and time >= self._lastVideoImage + self.VIDEO_PERIOD:
            self._lastVideoImage = time
            self._toVideo()
    
    def checkRecordingOnce(self):
        if self._recordOnce:
            self._record = False
            self._recordOnce = False
    
    def recordOnce(self):
        self._images = []
        self._record = True
        self._recordOnce = True
        self._lastVideoImage = 0.

    def record(self, record=True):
        if record:
            self._images = []
        self._record = record
        self._recordOnce = False
        self._lastVideoImage = 0.


class DrawOptions(object):
    def __init__(self, graphicalEngine):
        self.graphicalEngine = graphicalEngine
    
    @property
    def environment(self):
        return self.graphicalEngine.engine.environment



'''
    File name: dataset.py
    Author: Alexandre Manoury
    Python Version: 3.6
'''

import numpy as np
import math
import random
from scipy.spatial import cKDTree
from scipy.spatial.distance import euclidean

from exlab.modular.module import Module
from exlab.interface.serializer import Serializable
from exlab.utils.io import parameter

from .data import *
from .space import Space, SpaceKind
from .event import InteractionEvent
from .multispace import MultiColSpace, MultiColDataSpace, MultiRowDataSpace
from dinos.representation.entity_manager import EntityManager, Entity


# import graphviz


class SpaceManager(Module, Serializable, EntityManager):
    def __init__(self, name, storesData=False, options={}, parent=None, entityCls=Entity):
        Module.__init__(self, parent=parent)
        Serializable.__init__(self)
        EntityManager.__init__(self, name, entityCls=entityCls)
        self.name = name
        self.spaces = []
        self.storesData = storesData
        self.options = options

        self.multiColSpaces = []
        self.multiRowSpaces = []

        self.events = {}

        self.computeSpaces()
    
    def _sid(self, serializer):
        return serializer.uid('spaceManager', self.name)

    def _serialize(self, serializer):
        dict_ = {}
        dict_.update(serializer.serialize(self, ['spaces', 'storesData']))
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(dict_.get('storesData'),
                      options=dict_.get('options', {}))
        return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        serializer.set('spaceManager', self)
        serializer.deserialize(dict_.get('spaces', []))
        super()._postDeserialize(dict_, serializer)

    def __repr__(self):
        return f'SpaceManager({len(self.spaces)} spaces) and {EntityManager.__repr__(self)}'
    
    @property
    def size(self):
        return len(self.spaces)
    
    # Triggered when new data are added
    def updated(self):
        pass

    # Spaces
    def registerSpace(self, space):
        if space not in self.spaces:
            self.spaces.append(space)
            self.computeSpaces()

    def computeSpaces(self):
        self.actionSpaces = self.getActionSpaces(self.spaces)
        self.actionExplorationSpaces = self.getActionExplorationSpaces(self.spaces)
        self.actionPrimitiveSpaces = self.getActionPrimitiveSpaces(self.spaces)
        self.outcomeSpaces = self.getOutcomeSpaces(self.spaces)
    
    # Space conversion
    def convertSpace(self, space, kind=None, toData=None):
        toData = parameter(toData, self.storesData)  # space.canStoreData()
        kind = parameter(kind, space.kind)

        relatedSpace = next((s for s in self.spaces if s.kind.value == kind.value and s.linkedTo(space)), None)
        if relatedSpace is not None:
            return relatedSpace

        if toData:
            return space.createDataSpace(self, kind)
        else:
            return space.createLinkedSpace(self, kind)

    def convertSpaces(self, spaces, kind=None, toData=None):
        return [self.convertSpace(space, kind=kind, toData=toData) for space in spaces]

    def convertData(self, data, kind=None, toData=None):
        return data.convertTo(self, kind=kind, toData=toData)

    # Multi Spaces
    def _multiSpace(self, spaces, list_, type_, orderWise=False):
        # Only 1 space -> return the space itself
        if len(spaces) == 1:
            return spaces[0]

        # Look for exisiting multi space
        if orderWise:
            r = [s for s in list_ if s.spaces == spaces]
        else:
            r = [s for s in list_ if set(s.spaces) == set(spaces)]
        if len(r) == 0:
            s = type_(self, spaces)
            list_.append(s)
            return s
        else:
            return r[0]

    def _multiSpaceWeighted(self, spaces, list_, type_, weight=None, orderWise=False, weightSpaces=None):
        space = self._multiSpace(spaces, list_, type_, orderWise=orderWise)

        if weight:
            space.clearSpaceWeight()
            for subspace in weightSpaces:
                space.spaceWeight(weight, subspace)

        return space

    def multiColSpace(self, spaces, canStoreData=None, weight=None):
        # Flatten the list of spaces
        flatSpaces = list(set([subSpace for space in spaces for subSpace in space.cols if space is not None]))

        if len(flatSpaces) == 1:
            return flatSpaces[0]

        if canStoreData is None:
            if flatSpaces:
                dataSpaces = list(set([space.canStoreData()
                                       for space in flatSpaces]))
                if len(dataSpaces) > 1:
                    raise Exception(
                        f"All spaces should be a DataSpace or none: {flatSpaces}")
                canStoreData = dataSpaces[0]
            else:
                canStoreData = self.storesData

        return self._multiSpaceWeighted(flatSpaces, list_=self.multiColSpaces,
                                        type_=MultiColDataSpace if canStoreData else MultiColSpace, weight=weight, weightSpaces=spaces)

    def multiRowSpace(self, spaces, canStoreData=None):
        # spaces = list(set([subSpace for space in spaces for subSpace in space]))
        # spaces = list(set([space for space in spaces if space is not None]))
        spaces = [space for space in spaces if space is not None]

        if not spaces:
            return
        if len(spaces) == 1:
            return spaces[0]

        return self._multiSpaceWeighted(spaces, list_=self.multiRowSpaces, type_=MultiRowDataSpace, orderWise=True)

    # List Spaces
    def space(self, index_name, kind=SpaceKind.BASIC):
        return next(s for s in self.spaces if (s.name == index_name and (s.kind.value == kind.value or s.native == s))
                    or s.id == index_name)

    def spaceSearch(self, property=None, kind=SpaceKind.BASIC):
        if property:
            return next((s for s in self.spaces if s.boundProperty == property and s.kind.value == kind.value), None)
        return None

    def getActionSpaces(self, spaces):
        return [s for s in spaces if s.controllable() and s.kind == SpaceKind.BASIC]

    def getActionExplorationSpaces(self, spaces):
        return [s for s in spaces if s.controllable() and s.kind == SpaceKind.BASIC
                and not s.noaction]

    def getActionPrimitiveSpaces(self, spaces):
        return [s for s in spaces if s.primitive() and s.kind.value == SpaceKind.BASIC.value and not s.noaction]

    def getOutcomeSpaces(self, spaces):
        return [s for s in spaces if s.observable()
                and s.kind.value == SpaceKind.BASIC.value]

    # Data
    def addEvent(self, event, cost=1.):
        """Add data to the dataset"""
        event = event.clone()
        event.addToSpaces(cost=cost)
        self.logger.debug(f'Adding point {event} to dataset {self}', tag='dataset')
        self.events[event.iteration] = event

    def actions(self):
        return [event.actions for event in self.events]

    # def eventFromId(self, iteration):
    #     register = self.events[iteration]
    #     event = InteractionEvent(iteration)
    #     event.actions = ActionList(Action(
    #         *(SingleAction(t, self.getData(t, v).tolist()) for t, v in register.actions)))
    #     event.outcomes = Observation(
    #         *(SingleObservation(t, self.getData(t, v).tolist()) for t, v in register.outcomes))
    #     return event

    def getData(self, space_id, data_id):
        return self.space(space_id).data[data_id]



import numpy as np
import logging

from exlab.interface.serializer import Serializable


class Test(Serializable):
    def __init__(self, name, space, id=None):
        self.space = space
        self.name = name
        self.id = id
        self.scene = None
        self.points = []

        # Callbacks
        self.preTest = None
        self.prePointSetupEpisode = True
        self.prePoint = None
        self.postTest = None
    
    def _sid(self, serializer):
        return serializer.uid('test', self.name)

    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['space', 'name', 'id', 'scene', 'points'])
        return dict_

    # @classmethod
    # def _deserialize(cls, dict_, options={}, obj=None):
    #     obj = obj if obj else cls(Serializer.deserialize(dict_['space'], options=options),
    #                               dict_.get('name'),  dict_.get('id'))

    #     # operations
    #     obj.points = Serializer.deserialize(
    #         dict_.get('points'), options=options)

    #     return obj

    def _bindId(self):
        assert self.scene is not None
        index = 0

        def name(index):
            return f'{self.space.boundProperty.reference()}-{index}'

        while name(index) in self.scene.testIds:
            index += 1
        self.id = name(index)
        self.scene.testIds[self.id] = self

    def __repr__(self):
        return f'Test \'{self.name}\' on {self.space} ({len(self.points)} point(s))'


class TestResult(Serializable):
    def __init__(self, test, iteration, results, method=''):
        self.test = test
        self.iteration = iteration
        # [(error, goal, reached)]
        self.results = results
        self.method = method
        errorsArray = np.array([r[0] for r in results])
        self.meanError = np.mean(errorsArray)
        self.meanQuadError = np.mean(errorsArray ** 2)
        self.std = np.sqrt(self.meanQuadError - self.meanError ** 2)

        # logging.info("Error: {} {} {} [Sum, Quad, Std]".format(
        #     meanError, meanQuadError, std))

    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['iteration', 'meanError', 'meanQuadError', 'std', 'results', 'method'],
                                     foreigns=['test'])
        return dict_
    
    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(serializer.deserialize(dict_.get('test')),
                      dict_.get('iteration'),
                      dict_.get('results'),
                      dict_.get('method', ''))

        return super()._deserialize(dict_, serializer, obj)

    # @classmethod
    # def _deserialize(cls, dict_, options={}, obj=None):
    #     obj = obj if obj else cls(Serializer.deserialize(dict_['test'], options=options),
    #                               dict_['iteration'],
    #                               dict_['meanError'],
    #                               dict_['meanQuadError'],
    #                               dict_['std'],
    #                               Serializer.deserialize(
    #                                   dict_['results'], options=options),
    #                               method=dict_['method'])

    #     return obj

    def __repr__(self):
        return f'Result Test {self.test.id} @t={self.iteration} µ={self.meanError} σ={self.std} ({len(self.results)} point(s))'

    def details(self):
        results = [f'-{i}: aimed for {goal} and reached {reached}, error: {error}'
                   for i, (error, goal, reached) in enumerate(self.results)]
        results = '\n'.join(results)
        return f'{self}:\n{results}'


class PointTest(Test):
    def __init__(self, name, space, pointValue, relative=None):
        super().__init__(name, space)
        self.points.append(space.goal(pointValue).setRelative(relative))


class PointsTest(Test):
    def __init__(self, name, space, pointValueList, relative=None):
        super().__init__(name, space)
        for value in pointValueList:
            self.points.append(space.goal(value).setRelative(relative))


class UniformGridTest(PointsTest):
    def __init__(self, name, space, boundaries, numberByAxis=4, relative=None):
        vectors = [np.linspace(bound[0], bound[1], numberByAxis)
                   for bound in boundaries]
        mesh = np.meshgrid(*vectors)
        flatten = [axis.flatten() for axis in mesh]
        pointValueList = np.array(flatten).T

        super().__init__(name, space, pointValueList=pointValueList, relative=relative)



import random
import numpy as np
from scipy.optimize import minimize
from scipy.spatial.distance import euclidean
import copy
import time

from dinos.utils.move import MoveConfig

from dinos.utils.maths import sigmoid, threshold, linearValue

from dinos.data.data import Action, ActionList
# from dinos.data.data import *
from dinos.data.space import Space
from dinos.data.path import ActionNotFound, Path

from .random import RandomStrategy


class AutonomousStrategy(RandomStrategy):
    """AutonomousStrategy exploration strategy. Used in SAGG-RIAC algorithm."""

    def __init__(self, agent, name=None, performer=None, planner=None, options={}):
        """
        environment EnvironmentV4: environment of the learning agent
        dataset Dataset: episodic memory of the agent using the strategy (needed for local exploration)
        name string: name of the given strategy
        options dict: options of the strategy
        verbose boolean: indicates the level of debug info
        """
        super().__init__(agent, name=name, planner=planner, performer=performer,
                         options=options)
        # Contains the following keys:
        #   'min_points': min number of iterations before enabling the use of local exploration
        #   'nb_iteration': number of iterations per learning episode
        self.randomThreshold = self.options.get(
            'randomThreshold', 0.1)  # 1 -> only random
        self.randomFirstPassStart = 1.
        self.randomFirstPassEnd = 0.4
        self.randomFirstPassNumber = 200
        self.randomFirstPassNumberMin = 100

    def _serialize(self, serializer):
        dict_ = super()._serialize(serializer)
        dict_.update(serializer.serialize(self, []))
        return dict_

    # @classmethod
    # def _deserialize(cls, dict_, agent, options=None, obj=None):
    #     obj = obj if obj else cls(agent, dict_.get(
    #         'name'), options=dict_.get('options', {}))
    #     obj = RandomStrategy._deserialize(
    #         dict_, agent, options=options, obj=obj)
    #     return obj

    def _preRun(self, config):
        super()._preRun(config)

    def _runIteration(self, config):
        if config.goal:
            for _ in range(2):
                if self.reachGoalContext(config):
                    if self.exploreGoal(config):
                        return
                else:
                    break
        self.testRandomAction(config)

    def exploreGoal(self, config):
        # assert config.exploitation is False
        assert config.depth == 0

        # for _ in self.performer.iterative():
        # Choose between local and global exploration
        config.absoluteGoal = config.goal.absoluteData(self.agent.environment.state())
        probUseRandom, path = self.useGoalCriteria(config.goal, config)
        useGoal = random.uniform(0, 1) > probUseRandom
        config.result.randomProbability = probUseRandom

        self.logger.debug(f'goal exploration decision: criteriaRandom={probUseRandom}->useGoal={useGoal} exploration={config.exploitation} path={path}')

        if useGoal:  # We have chosen local optimizattion
            self.testPath(path, config)
            config.result.reachedGoal = self.agent.environment.state().context().projection(config.goal.space)
            config.result.reachedGoalDistance = (config.result.reachedGoal - config.absoluteGoal).norm()
            config.result.reachedGoalStatus = 'goal attempted'
        elif not config.exploitation:  # We have chosen random exploration
            if config.goalContext and not config.changeContext:
                config.changeContext = True
                return False
            else:
                self.testRandomAction(config)
                config.result.reachedGoalStatus = 'random chosen instead'
        else:
            self.testRandomAction(config, zero=True)
            config.result.reachedGoalStatus = 'zero chosen because exploitation'

        return True

    def useGoalCriteria(self, goal, config):
        """Criteria used to choose between local and global exploration (Straight from Matlab SGIM code)."""
        prob = 1.0

        # First pass: only random
        if goal.space.number < self.randomFirstPassNumberMin:
            return 30., Path()

        probFirstPass = linearValue(self.randomFirstPassStart, self.randomFirstPassEnd,
                                    (goal.space.number - self.randomFirstPassNumberMin) / self.randomFirstPassNumber)
        # print(probFirstPass)
        if not config.exploitation and random.uniform(0, 1) < probFirstPass:
            return 20. + probFirstPass, Path()  # Random action

        # Try planning to goal
        try:
            path, _, distance = self.planner.planDistance(goal, model=config.model, settings=config.plannerSettings)
        except ActionNotFound:
            path = Path()
        # print(path)
        if not path:
            config.result.reachedGoal = 'planning failed'
            self.logger.warning(
                f"Planning failed for goal {goal}, switching to random")
            return 10., Path()  # Random action

        # Compute criteria
        if config.exploitation:
            prob = -1.
        else:
            error = distance / (goal.space.maxDistance * 0.1)
            # (distance - space.options['err']) / space.options['range']
            # prob = 0.8*(sigmoid(error) - 0.5) + 0.5
            # prob = (prob - self.randomFirstPass) / (1. - self.randomFirstPass)
            prob = 0.1 + 0.9 * error  # threshold(prob, self.randomThreshold)

        '''space = goal.space
        if len(space.data) > self.options['min_points']:
            _, dist = space.nearest(goal, 5)
            if len(dist) < 5:
                return prob
            mean_dist = np.mean(dist)
            x = (mean_dist - space.options['err']) / space.options['range']
            prob = 0.8*(sigmoid(x) - 0.5) + 0.5'''
        # prob = 0.9*(sigmoid(x) - 0.5) + 0.5   # From matlab SGIM code
        return prob, path

    # def runLocalOptimization(self, a, goal, initial_simplex=None, config=None):
    #     """Perform local exploration using Nelder-Mead optimization method."""

    #     # Function to minimize for Nelder-Mead
    #     # if config.exploitation else (self.options['nb_iteration'] - self.n) # TODO  id:10
    #     maxiter = 1
    #     aspace = a.get_space()

    #     def function(x):
    #         return self.testNelderMead(x, aspace, goal, config)

    #     minimize(function, a.flatten(), method=custom_nelder_mead, options={'xtol': 1e-3, 'ftol': 1e-6,
    #                                                                         'maxiter': maxiter, 'maxfev': maxiter,
    #                                                                         'initial_simplex': initial_simplex})

    # def testNelderMead(self, a_data, aspace, g, config):
    #     """Method called when testing an action with Nelder-Mead algorithm."""

    #     a = ActionList(aspace.asTemplate(a_data.tolist()))

    #     self.test_action(a, config)
    #     space = g.get_space()
    #     dist = space.options['out_dist']/space.options['max_dist']
    #     '''if not config.exploitation:
    #         for i in range(len(self.memory[-1][-1][3])):
    #             if self.memory[-1][-1][3][i] == space.id:
    #                 dist = min(euclidean(g, self.memory[-1][-1][2][i]), space.options['out_dist']) / space.options['max_dist']'''
    #     return dist



'''
    File name: objects.py
    Author: Alexandre Manoury
    Python Version: 3.6
'''

import copy
import random

from exlab.interface.serializer import Serializable
from exlab.utils.io import parameter

from dinos.agents.tools.planners.planner import PlanSettings

from .result import Result


"""
Misc objects
"""


class MoveConfig(Serializable):
    """
    Represents a move configuration:
    - Exploitation:
        * Model
        * Goal
        * (recursion depth)
    - Training:
        - Goal oriented:
            * Model
            * Goal
            * Strategy
            * (recursion depth)
        - Action oriented:
            * Strategy
            * (recursion depth)
    """

    def __init__(self, model=None, exploitation=False, depth=0, strategy=None, goal=None, goalContext=None,
                 changeContext=False, lastEvent=None, sampling='', iterations=1, iteration=-1, allowReplanning=True,
                 evaluating=False, plannerSettings=None):
        self.exploitation = exploitation
        self.evaluating = evaluating
        self.depth = depth

        self.strategy = strategy
        self.model = model
        self.goal = goal
        self.goalContext = goalContext
        self.changeContext = changeContext
        self.absoluteGoal = None
        self.sampling = sampling

        self.allowReplanning = allowReplanning
        self.plannerSettings = parameter(plannerSettings, PlanSettings())

        self.iterations = iterations
        self.iteration = iteration

        self.result = Result(self)
        self.plannerSettings.result = self.result
    
    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['exploitation', 'evaluating', 'depth', 'goal', 'goalContext', 'absoluteGoal',
                                            'sampling', 'allowReplanning', 'iterations', 'iteration', 'result'],
                                     foreigns=['strategy', 'model'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls()
        return super()._deserialize(dict_, serializer, obj)
    
    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        for attr in ['exploitation', 'evaluating', 'depth', 'goal', 'goalContext', 'absoluteGoal',
                     'sampling', 'allowReplanning', 'iterations', 'iteration', 'strategy', 'model']:
            if attr in dict_:
                setattr(self, attr, serializer.deserialize(dict_.get(attr)))
        if 'result' in dict_:
            serializer.deserialize(dict_.get('result'), obj=self.result)

    def clone(self, **kwargs):
        new = copy.copy(self)
        for key, value in kwargs.items():
            setattr(new, key, value)
        
        new.result = new.result.clone(new)
        return new

    def nextdepth(self, **kwargs):
        kwargs['depth'] = self.depth + 1
        return self.clone(**kwargs)
    
    @property
    def training(self):
        return not self.evaluating and not self.exploitation

    def __repr__(self):
        if self.evaluating:
            prefix = "Evaluation"
            attrs = ['model', 'goal', 'absoluteGoal', 'depth']
        elif self.exploitation:
            prefix = "Exploit"
            attrs = ['model', 'goal', 'absoluteGoal', 'depth']
        elif self.goal:
            prefix = "Goal exploration"
            attrs = ['goal', 'absoluteGoal', 'goalContext',
                     'model', 'strategy', 'depth', 'sampling']
        else:
            prefix = "Action exploration"
            attrs = ['strategy', 'depth']
        params = ', '.join([f'{k}: {getattr(self, k)}' for k in attrs])
        return f'Config ({prefix}) [{params}] [{self.result}]'



'''
    File name: performer.py
    Author: Alexandre Manoury
    Python Version: 3.6
'''

from exlab.modular.module import Module

from dinos.utils.move import MoveConfig

from dinos.data.event import InteractionEvent
from dinos.data.data import ActionList, Data
from dinos.data.space import SpaceKind, FormatParameters
from dinos.data.path import Path, ActionNotFound

from dinos.agents.tools.planners.planner import PlanSettings
from dinos.agents.tools.policies.policy import LearningPolicy


class Performer(Module):
    """
    Executes Paths created by a planner
    """

    def __init__(self, agent, options={}):
        super().__init__('Performer', agent, loggerTag='performer')

        self.agent = agent
        self.environment = self.agent.environment
        self.options = options

        self.iterations = max(1, options.get('iterations', 1))
        self.lastSequence = None

    def perform(self, path, config=MoveConfig()):
        """
        Tests a specific Path and stores consequences in memory.
        :param path: Path object (created by a planner)
        """
        r = self.__perform(path, config)[0]
        return r

    def performActions(self, actionList, goal=None, config=MoveConfig()):
        """
        Tests a specific action list and stores consequences in memory.
        :param actionList: ActionList
        """
        path = self.agent.planner.planActions(actionList)
        return self.__perform(path, config)[0]

    def __perform(self, path, config, i=0, depth=0):
        """
        Tests a specific action list and stores consequences in memory.
        """
        result, p = Performance.perform(self, path, config, i=i, depth=depth)
        self.lastSequence = p.lastSequence
        return result

    def iterative(self):
        return range(self.iterations)


class PerformedInfo(object):
    def __init__(self, message):
        self.message = message
    
    def __repr__(self):
        return self.message


class PerformedSequence(object):
    def __init__(self, node, depth):
        self.node = node
        self.depth = depth

        self.startingPosition = None
        self.startingDerive = None
        self.endingPosition = None
        self.endingDerive = None

        self.predictedOutcome = None
        self.startingPredictionDerive = None
        self.endingPredictionDerive = None
    
    def __repr__(self):
        spaces = '  ' * self.depth
        return f'{spaces} Prediction({self.predictedOutcome} estimated ±{self.startingPredictionDerive:.4f} and got ±{self.endingPredictionDerive:.4f})  \
Reality({self.startingPosition.plain()} ±{self.startingDerive:.4f})->({self.endingPosition.plain()} ±{self.endingDerive:.4f})'
    
    @property
    def action(self):
        return self.node.action
    
    @property
    def startingWantedPosition(self):
        return self.node.parent.absPos
    
    @property
    def endingWantedPosition(self):
        return self.node.absPos


class Performance(object):
    MAX_DERIVE = 0.04
    MAX_DISTANCE = 0.01
    RECOMPUTE_EXECUTION = False
    MAX_PREDICTED_LINEAR_ERROR = 0.002

    def __init__(self, performer, path, config, parent=None, i=0, depth=0):
        self.performer = performer
        self.path = path
        self.config = config
        self.depth = depth
        self.i = i

        self.agent = self.performer.agent
        self.environment = self.performer.environment
        self.logger = self.performer.logger

        self.parent = parent
        self.root = self.parent.root if self.parent is not None else self

        self.initVariables()
    
    @classmethod
    def perform(self, performer, path, config, parent=None, i=0, depth=0):
        p = Performance(performer, path, config, parent, i, depth)
        return p.execute(), p
    
    def initVariables(self):
        self.results = []
        self.lastResults = []
        self.lastSequence = []
        self.lastSequenceNode = None
        # print('((START))')
    
        self.absoluteGoal = None
        if self.path.goal:
            self.absoluteGoal = self.path.goal.absoluteData(self.environment.state())
        if self.path.planSettings:
            self.path.planSettings.performing = True
        self.model = self.path.model

        self.replanning = 0
        self.maxReplanning = 5
        self.maxDistance = 7.
        self.maxDerive = None
        if self.model:
            self.maxDistance = self.model.getPrecision(self.MAX_DISTANCE, 1.5)
            self.maxDerive = self.model.getPrecision(self.MAX_DERIVE, 2.5)
        self.nodes = list(self.path.nodes)

        self.formatParameters = FormatParameters()
        self.observationsPrevious = None
        self.distanceToGoal = None
        self.distanceToGoalSinglePass = None
        self.currentPos = None

        self.success = False
        self.fatal = False
        self.topReplanning = False
    
    def execute(self):
        while not self.success and not self.fatal:
            if self.replan():
                break

            if not self.nodes:
                break

            for node in self.nodes:
                self.lastSequenceNode = PerformedSequence(node, self.depth)
                self.root.lastSequence.append(self.lastSequenceNode)

                if self.performNodeContext(node):
                    break

                if self.computePrimitiveForNode(node):
                    break

                if self.checkPreExecution(node):
                    break

                if self.performNode(node):
                    break

                # self.logger.debug(f'Iter (d{self.depth}) {self.i}:\n{node.ty0} Estimated state:\n{node.state.context()}\nCurrent New State:\n{self.environment.state().context()}')
                # self.logger.info(f'Iter (d{self.depth}) {self.i}: performing {node.action}')

                self.processMemory()
                self.checkPostExecution(node)

                if self.checkGoalReached(node):
                    break

                if self.checkDerive(node):
                    break
                self.i += 1
            self.nodes = None
        # if o:
        #     y = o.difference(oStart)
        #results.append(InteractionEvent(self.getIteration(), actionListExecuted, y))

        # print('((OVER))')

        self.postPerformance()

        return self.results, (self.fatal, self.topReplanning)

    def replan(self):
        if not self.nodes and self.absoluteGoal and self.config.allowReplanning:
            if self.distanceToGoal:
                if self.replanning == 0:
                    self.distanceToGoalSinglePass = self.distanceToGoal
                # self.postPerforming(model, False, distanceToGoal)
            self.replanning += 1
            self.config.result.performerReplanning += 1
            if self.replanning > self.maxReplanning:
                self.logger.warning(f'Iter (d{self.depth}) {self.i}: out of replanning')
                return True
            try:
                self.logger.warning(f'Iter (d{self.depth}) {self.i}: no more action to execute... trying to replan new ones')
                newPath, _ = self.agent.planner.plan(self.absoluteGoal, settings=self.path.planSettings)
                if newPath:
                    self.nodes = newPath.nodes
            except ActionNotFound:
                self.logger.warning(f'Iter (d{self.depth}) {self.i}: failed to plan new one')
        return False
    
    def performNodeContext(self, node):
        if node.context:
            subResults, (self.fatal, self.topReplanning) = self.perform(self.performer, node.context, self.config, parent=self, depth=self.depth+1)[0]
            if self.fatal:
                return True
            self.results += subResults
            self.observationsPrevious = None
        return False
    
    def performNode(self, node):
        if node.execution:
            self.observationsPrevious = self.agent.observe(formatParameters=self.formatParameters)
            subResults, (self.fatal, self.topReplanning) = self.perform(self.performer, node.execution, self.config, parent=self, depth=self.depth+1)[0]
            if self.fatal:
                return True
            self.results += subResults
            observations = self.agent.observe(formatParameters=self.formatParameters)
            self.differences = observations.difference(self.observationsPrevious)
            self.observationsPrevious = None
        else:
            # self.logger.debug(f'Iter (d{self.depth}) {self.i}:\n{node.model.npForward(node.action, currentState.context())}')
            event, self.differences, self.observationsPrevious = self.performAction(node, self.observationsPrevious)
            self.results.append(event)
            self.lastResults.append(event)
        return False
    
    def computePrimitiveForNode(self, node):
        if not node.action.space.primitive() and (self.RECOMPUTE_EXECUTION or not node.execution):
            try:
                node.execution, _ = self.agent.planner.plan(node.action, settings=PlanSettings(performing=True))
            except ActionNotFound:
                pass
            if not node.execution:
                self.logger.warning(f'Iter (d{self.depth}) {self.i}: failed to break down non primitive action {node.action}')
                self.fatal = True
                return True
        return False
    
    def postPerforming(self, success, distanceToGoal, singlePrecision=False):
        if self.model:
            self.model.updatePrecision(success, distanceToGoal, index=int(singlePrecision))
    
    def postPerformance(self):
        if self.distanceToGoal:
            if self.replanning == 0:
                self.distanceToGoalSinglePass = self.distanceToGoal
            if self.distanceToGoalSinglePass:
                self.postPerforming(self.success, self.distanceToGoalSinglePass, True)
            self.postPerforming(self.success, self.distanceToGoal * (1 + self.replanning * 0.1))

            if self.depth == 0:
                self.config.result.performerDistance = (self.distanceToGoal, self.maxDistance)
    
    def performAction(self, node, observationsPrevious):
        rawAction = node.action
        action = (node.goal if node and node.goal else rawAction).clone()
        primitiveAction = rawAction

        if observationsPrevious is None:
            observationsPrevious = self.agent.observe(formatParameters=self.formatParameters)
        self.agent._performAction(rawAction, config=self.config)
        observations = self.agent.observe(formatParameters=self.formatParameters)
        differences = observations.difference(observationsPrevious)
        event = InteractionEvent(self.environment.counter.last,
                                 action,
                                 primitiveAction,
                                 differences,
                                 observationsPrevious.convertTo(kind=SpaceKind.PRE))
        observationsPrevious = observations

        return event, differences, observationsPrevious
    
    def processMemory(self):
        if self.agent.learningPolicy == LearningPolicy.EACH_ITERATION:
            self.agent.addMemory(self.lastResults, self.config)
            self.lastResults = []

    def checkGoalReached(self, node):
        if self.absoluteGoal:
            relative = self.absoluteGoal.relativeData(self.currentState)
            self.distanceToGoal = relative.norm()
            if self.distanceToGoal < self.maxDistance and node == self.nodes[-1]:
                self.root.lastSequence.append(PerformedInfo(f'Goal reached {self.distanceToGoal:.3f} < {self.maxDistance:.3f}'))
                self.logger.debug(f'Iter (d{self.depth}) {self.i}: close enough to goal! {self.distanceToGoal:.3f} (max {self.maxDistance:.3f}) ({relative})')
                self.success = True
                return True
            else:
                self.logger.debug(f'Iter (d{self.depth}) {self.i}: distance to goal {self.distanceToGoal:.3f} (max {self.maxDistance:.3f}) ({relative})')
        return False
    
    def checkStatus(self, node, goal, currentState=None):
        if currentState is None:
            currentState = self.environment.state()
        derive = goal.relativeData(currentState, ignoreRelative=True)
        currentPos = goal - derive
        distanceDerive = derive.norm()
        return currentPos, distanceDerive, currentState
    
    def checkPreExecution(self, node):
        if node.absPos:
            self.currentPos, derive, self.currentState = self.checkStatus(node, node.parent.absPos)
            self.logger.debug(
                f'Iter (d{self.depth}) {self.i}: pre execution check: should be at {node.parent.absPos} and currently at {self.currentPos}, ' +
                f'diff {derive:.4f}\nCurrent Pre State:\n{self.currentState.context()}')
            
            self.lastSequenceNode.startingPosition = self.currentPos
            self.lastSequenceNode.startingDerive = derive
            
            if self.model:
                context = self.environment.state().context()
                # self.logger.info(f'Iter (d{self.depth}) {self.i}: {context}')
                predictedOutcome, predictedLinearError = self.model.forward(node.action, context=context)
                predictionDerive = (predictedOutcome - node.goal).norm()

                self.lastSequenceNode.predictedOutcome = predictedOutcome
                self.lastSequenceNode.startingPredictionDerive = predictionDerive

                # if predictionDerive > self.maxDerive * .5 or 
                if predictedLinearError > self.MAX_PREDICTED_LINEAR_ERROR:
                    self.root.lastSequence.append(PerformedInfo(
                        f'Replanning because of exceeding prediction derive {predictionDerive:.4f} > {self.maxDerive:.4f} ({predictedLinearError:.6f})'))
                    self.logger.info(
                        f'Iter (d{self.depth}) {self.i}: max prediction derive exceeded ({predictionDerive:.4f} > {self.maxDerive:.4f}) trying to reach {node.goal}')

                    if self.replanning <= self.maxReplanning:
                        self.logger.info(f'Replanning...')
                        self.nodes = None
                        return True
                    else:
                        self.logger.warning(f'No replanning left!')

                    self.root.lastSequence.append(PerformedInfo(f'Replanning because of exceeding prediction derive {predictionDerive:.4f} > {self.maxDerive:.4f}'))
        return False
    
    def checkPostExecution(self, node):
        self.currentState = self.environment.state()
        # self.logger.warning(currentState)
        self.relativeDifference, self.relativeDerive = None, -1
        if node.absPos:
            self.currentPos, derive, self.currentState = self.checkStatus(
                node, node.absPos, self.currentState)

            # if differences is not None else None
            self.relativeDifference = self.differences.projection(node.absPos.space)
            # if differences is not None else -1.
            self.relativeDerive = (self.relativeDifference - node.goal).norm()
            if self.model:
                self.model.updatePrecisionPerGoal(node.goal, self.relativeDerive)
            self.logger.debug(
                f'Iter (d{self.depth}) {self.i}: wanting {node.absPos} and got {self.currentPos} \n     Diff {derive:.4f} doing {node.action} to get {node.goal} ' +
                f'and got {self.relativeDifference} Diff {self.relativeDerive:.4f}')
            
            if self.model:
                self.lastSequenceNode.endingPredictionDerive = (self.lastSequenceNode.predictedOutcome - self.relativeDifference).norm()
            
            self.lastSequenceNode.endingPosition = self.currentPos
            self.lastSequenceNode.endingDerive = self.relativeDerive

    
    def checkDerive(self, node):
        if node.absPos and self.maxDerive:
            self.currentPos, derive, self.currentState = self.checkStatus(node, node.absPos, self.currentState)
            # print(node.action)
            # print(derive, maxDerive)
            if derive > self.maxDerive:
                if self.model:
                    predictedOutcome, _ = self.model.forward(node.action, context=self.environment.state().context())
                    self.logger.info(f'Iter (d{self.depth}) {self.i}: prediction is {predictedOutcome} ?= planned {node.goal} ?= got {self.relativeDifference}')
                self.root.lastSequence.append(PerformedInfo(f'Replanning because of exceeding derive {derive:.4f} > {self.maxDerive:.4f}'))
                self.logger.info(
                    f'Iter (d{self.depth}) {self.i}: max derive exceeded ({derive:.4f} > {self.maxDerive:.4f}) trying to reach {node.goal} by doing ' +
                    f'{node.action} to get {node.goal} and got {self.relativeDifference} Diff {self.relativeDerive:.4f}')

                self.postPerforming(False, derive)
                self.postPerforming(False, derive, True)

                if self.depth == 0:
                    self.config.result.performerDerive.append((derive, self.maxDerive))
                if self.replanning <= self.maxReplanning:
                    self.logger.info(f'Replanning...')
                    self.nodes = None
                    return True
                else:
                    self.logger.warning(f'No replanning left!')
            # print(f'Derive is {derive} {node.absPos} {node.absPos.relativeData(self.environment.state(), ignoreRelative=True)}')
            # print(self.environment.state().context())
            # print(f'Max {self.MAX_DERIVE * node.absPos.space.maxDistance}')
        return False





'''
    File name: space.py
    Author: Alexandre Manoury, Nicolas Duminy
    Python Version: 3.6
'''

import sys
import copy
import math
import random
import logging
import numpy as np

from enum import Enum
from scipy.spatial.distance import euclidean
from sklearn.neighbors import NearestNeighbors

from exlab.interface.serializer import Serializable
from exlab.utils.io import parameter

# from dinos.utils.io import getVisual, plotData, visualize
# from dinos.utils.logging import Logger
from dinos.utils.maths import popn
from dinos.data.data import SingleData, Data, Goal, Action


class SpaceKind(Enum):
    # NATIVE = 'native'
    BASIC = 'basic'
    PRE = 'pre'


"""
"""


class FormatParameters(object):
    def __init__(self):
        self.spaces = {}


class Space(Serializable):
    RESERVE_STEP = 10000
    _number = 0

    def __init__(self, spaceManager, dim, options={}, native=None, kind=SpaceKind.BASIC, spaces=None, property=None):
        self.id = Space._number
        Space._number += 1

        self.spaces = spaces if spaces is not None else [self]
        self.rows = self.spaces
        self.aggregation = len(self.spaces) > 1
        self.rowAggregation = False
        self.childrenSpaces = []

        # self.allowedSimilarRows = []

        self.native = native if native else self
        self.nativeRoot = self
        self.nativeRoot = self.native.nativeRoot
        self.kind = kind
        self._property = property

        self.dim = dim
        self.options = options
        # self.delegateto = options.get('delegateto', None)
        self._relative = options.get('relative', False)
        self._relativeLearning = options.get('relativeLearning', True)
        self._modulo = options.get('modulo', None)
        self.noaction = options.get('noaction', False)

        self.abstract = False
        if not self.aggregation:
            self.abstract = any(s.abstract for s in self.spaces)

        self._bounds = [[-1., 1.] for i in range(self.dim)]
        self.maxDistance = 1.
        self.maxDistancePerColumn = np.full(self.dim, 1.)

        # Multi
        self.invalid = False
        self.invalidate()

        # Register
        self.spaceManager = spaceManager
        self.spaceManager.registerSpace(self)
    
    def _sid(self, serializer):
        return serializer.serialize(self, ['kind'], foreigns=['spaceManager', 'boundProperty'], reference=True)

    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['kind', 'options'], foreigns=['spaceManager', 'boundProperty'])
        return dict_
    
    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = serializer.get('environment').propertySpace(serializer.deserialize(dict_.get('boundProperty')),
                                                              SpaceKind(dict_.get('kind')),
                                                              serializer.deserialize(dict_.get('spaceManager')))
        return super()._deserialize(dict_, serializer, obj)
    
    # def _postDeserialize(self, dict_, serializer):
    #     super()._postDeserialize(dict_, serializer)

    # @classmethod
    # def _deserialize(cls, dict_, spaceManager, options=None, obj=None):
    #     # Setting gamma
    #     # deprecated
    #     options = dict_.get('options', {})
    #     if 'gamma' in options.keys():
    #         cls.gamma = options['gamma']

    #     # Creating object
    #     spaces = None
    #     obj = obj if obj else cls(spaceManager, dict_.get('dimension', 1), dict_.get('options', {}),
    #                               native=dict_.get('native'), kind=dict_.get('kind', SpaceKind.BASIC),
    #                               spaces=spaces)

    #     # Operations
    #     # Loading results
    #     # if options.get('loadResults') and dict_.get('_number', -1) >= 0:
    #     #     obj.id = dict_.get('id', -1)  # TODO check for id collision id:21
    #     #     obj._number = dict_.get('_number', 0)
    #     #     obj.actions = dict_.get('actions', obj.actions)
    #     #     obj.data = dict_.get('data', obj.data)
    #     #     obj.costs = dict_.get('costs', obj.costs)
    #     #     obj.ids = dict_.get('ids', obj.ids)
    #     #     obj.lids = dict_.get('lids', obj.lids)
    #     return obj

    # Properties
    # def nativeRoot(self):
    #     if self.native == self:
    #         return self
    #     return self.native.nativeRoot()
    
    @property
    def boundProperty(self):
        if self._property:
            return self._property
        return self.nativeRoot._property
    
    def observable(self):
        if self.boundProperty is None or self.null():
            return False
        return self.boundProperty.observable()

    def controllable(self):
        return True

    def primitive(self):
        if self.boundProperty is None or self.null():
            return False
        return self.boundProperty.controllable()
    
    def canStoreData(self):
        return False
    
    @property
    def learnable(self):
        prop = self.boundProperty
        if prop:
            return prop.learnable
        return False

    @property
    def relative(self):
        return self.nativeRoot._relative
    
    @property
    def relativeLearning(self):
        return self.nativeRoot._relativeLearning

    @property
    def modulo(self):
        return self.nativeRoot._modulo

    def null(self):
        return not self.spaces
    
    def __bool__(self):
        return not self.null()

    # Attributes
    @property
    def cols(self):
        return self.spaces

    # @property
    # def rows(self):
    #     return self.spaces + [row for row in self.allowedSimilarRows if row not in self]

    def allowSimilarRows(self, rows):
        self.rows += [row for row in rows if row not in self]
    
    def resetSimilarRows(self):
        self.rows = self.spaces

    @property
    def baseCols(self):
        return self.spaces

    @property
    def baseRows(self):
        return [self.spaces[0]]

    @property
    def flatColsWithoutMultiRows(self):
        # if not self.spaces:
        #     return []
        if not self.aggregation:
            return self.baseCols
        return [colSpace for space in self.baseCols for colSpace in space.flatColsWithoutMultiRows]

    @property
    def flatColsWithMultiRows(self):
        if len(self.cols) == 1 and self.cols[0] == self:
            return self.cols
        return [colSpace for space in self.cols for colSpace in space.flatColsWithMultiRows]

    @property
    def flatSpaces(self):
        if not self.aggregation:
            return self.spaces
        return [colSpace for space in self.spaces for colSpace in space.flatSpaces]

    @property
    def groupedCols(self):
        if not self.aggregation:
            return [([self],)]
        return [colSpace for space in self.baseCols for colSpace in space.groupedCols]

    # @property
    # # List all col-stacked (Vertical Stack) spaces in the current space
    # def colSpaces(self):
    #     return self.spaces

    # @property
    # def colSpacesAll(self):
    #     if not self.spaces:
    #         return []
    #     if self.spaces[0] == self:
    #         return self.spaces
    #     return [colSpace for space in self.spaces for colSpace in space.colSpacesAll]

    # @property
    # def colSpacesFull(self):
    #     return self.spaces

    # @property
    # # List all row-stacked (Horizontal Stack) spaces
    # def rowSpaces(self):
    #     return [self.spaces[0]]

    def columnsFor(self, space):
        pos = 0
        colIds = []
        for s in self.cols:
            if s.intersects(space):
                colIds += range(pos, pos + s.dim)
            pos += s.dim
        return colIds
    
    def findMatchingSpaceRows(self, other, kindSensitive=True, idSensitive=False, entity=None):
        # print(self.flatColsWithMultiRows)
        # print(self.flatColsWithoutMultiRows)
        # print(other.flatColsWithMultiRows)
        # print(other.flatColsWithoutMultiRows)
        for row in self.rows:
            for otherRow in other.rows:
                if row.matches(otherRow, kindSensitive=kindSensitive, idSensitive=idSensitive, entity=entity):
                    return (row, otherRow)
        return None

    def matches(self, other, kindSensitive=True, dataSensitive=False, idSensitive=False, rowMatching=False, entity=None):
        if idSensitive and self != other:
            return False
        if entity and (not self.matchesEntity(entity) or not other.matchesEntity(entity)):
            return False
        if kindSensitive and self.kind.value != other.kind.value:
            return False
        if dataSensitive and self.canStoreData != other.canStoreData:
            return False
        if rowMatching and self.findMatchingSpaceRows(other, kindSensitive=kindSensitive, idSensitive=idSensitive):
            return True
        return self.nativeRoot == other.nativeRoot

    def linkedTo(self, other):
        return self.nativeRoot == other.nativeRoot
    
    def intersects(self, space):
        if isinstance(space, list):
            space = set([sp for s in space for sp in s.flatSpaces])
        else:
            space = set(space.flatSpaces)
        return set(self.flatSpaces).intersection(space)
    
    def includes(self, space):
        return space in self.rows

    # Multi
    def __iter__(self):
        return self.baseCols.__iter__()

    # Points
    def point(self, value, relative=None, toSpace=None):
        d = Data(toSpace if toSpace is not None else self, value)
        d.setRelative(relative)
        return d

    def goal(self, value, relative=None):
        d = Goal(self, value)
        d.setRelative(relative)
        return d

    def action(self, value, relative=None):
        d = Action(self, value)
        d.setRelative(relative)
        return d

    def zero(self, relative=None):
        self._validate()
        d = Data(self, [0.] * self.dim)
        d.setRelative(relative)
        return d

    def plainZero(self):
        self._validate()
        return np.array([0.] * self.dim)
    
    def randomPoint(self, relative=None):
        d = Action(self, [random.uniform(minb, maxb) for minb, maxb in self.bounds])
        d.setRelative(relative)
        return d

    def plainRandomPoint(self):
        return np.array([random.uniform(minb, maxb) for minb, maxb in self.bounds])

    def asTemplate(self, data, type_item=SingleData, type_vector=Data, entity=None):
        data = list(data)
        if len(data) != self.dim:
            logging.critical(f"Template dimension mismatch: space {self.name} is {self.dim}d and data is {len(data)}d")
        if entity and len(list(self)) == 1:
            parts = [type_item(s, data)
                     for s in self.rows if s.matchesEntity(entity)]
            if not parts:
                parts = [type_item(s, data) for s in self]
        else:
            parts = [type_item(s, popn(data, s.dim)) for s in self]
        return type_vector(*parts)

    def formatData(self, data, formatParameters=None):
        if type(data) is not np.ndarray:
            data = np.array(data)
        if formatParameters:
            if self.native not in formatParameters.spaces:
                formatParameters.spaces[self.native] = {}
            settings = formatParameters.spaces[self.native]
        else:
            settings = {}

        if self.modulo is not None:
            settings['modulo'] = settings.get('modulo', data // self.modulo)
            data = data - settings['modulo'] * self.modulo

        return data
    
    def matchesEntity(self, entity):
        if not entity:
            return True
        return self.boundProperty.entity == entity
    
    def applyTo(self, entity):
        if not entity or len(self.rows) == 1:
            return self
        return next((row for row in self.rows if row.matchesEntity(entity)), self)

    # Data
    @property
    def bounds(self):
        self._validate()
        return copy.deepcopy(self._bounds)

    @staticmethod
    def infiniteBounds(dim):
        return [[-math.inf, math.inf] for i in range(dim)]
    
    def convertTo(self, spaceManager=None, kind=None, toData=None):
        spaceManager = parameter(spaceManager, self.spaceManager)
        return spaceManager.convertSpace(self, kind=kind, toData=toData)

    def createLinkedSpace(self, spaceManager=None, kind=None):
        kind = parameter(kind, self.kind)
        spaceManager = spaceManager if spaceManager else self.spaceManager
        return Space(spaceManager, self.dim, native=self.native, kind=kind)

    def createDataSpace(self, spaceManager=None, kind=None):
        from .dataspace import DataSpace
        kind = parameter(kind, self.kind)
        spaceManager = spaceManager if spaceManager else self.spaceManager
        return DataSpace(spaceManager, self.dim, native=self.native, kind=kind)

    # Validation
    def invalidate(self):
        if self.invalid:
            return

        self.invalid = True
        for space in self.childrenSpaces:
            space.invalidate()

    def _validate(self):
        if self._preValidate():
            self._postValidate()

    def _preValidate(self):
        if not self.invalid:
            return False
        if len(self.spaces) > 1:
            for space in self.spaces:
                space._validate()
        return True

    def _postValidate(self):
        self.invalid = False
    
    # Representation
    @property
    def name(self):
        return self.boundProperty.absoluteName if self.boundProperty else None

    def icon(self):
        return '@'

    def colStr(self):
        return self.boundedProperty()
    
    def boundedProperty(self):
        if self._property:
            return f"→{self._property}"
        elif self.boundProperty:
            return f"↝{self.boundProperty}"
        else:
            return "↛"

    def toStr(self, short=False):
        if not self.spaces:
            return '@NullSpace'
        absName = f'#{self.name}' if self.name else ''
        suffix = '' if self.kind.value == SpaceKind.BASIC.value else f':{self.kind.value.upper()}'
        if short == 2:
            return f"{self.boundedProperty()}"
        if short:
            return f"#{self.id}{absName}{self.colStr()}{suffix}↕{self.dim} {self.icon()}"
        return f"{self.icon()}#{self.id}{absName}{self.colStr()}{suffix}↕{self.dim}"

    def __repr__(self):
        return self.toStr()
    
    # Deprecated
    def iterate(self):
        return self.baseCols



'''
    Author: Alexandre Manoury
    Python Version: 3.6
'''

import numpy as np

from exlab.interface.serializer import Serializable, Serializer
from exlab.utils.text import colorText, Colors
from exlab.utils.ensemble import Ensemble
from exlab.utils.io import parameter

from dinos.data.data import Observation
from dinos.data.space import SpaceKind


class Entity(Serializable):
    """Represents a world entity with a set of properties (features and effectors)

    Args:
        kind (string): the type of our entity
        absoluteName (string): Absolute name used to point to this entity. Unique

    Attributes:
        index (int): Absolute index
        indexKind (int): Index for the given entity type
        parent (Entity):
        activated (bool):
        absoluteName
        kind

    """

    number = 0
    indexes = {}

    def __init__(self, kind, absoluteName='', disconnected=False, manager=None):
        self.absoluteName = absoluteName
        self.kind = kind

        # self.proxy = proxy

        self._manager = manager

        self.disconnected = disconnected
        self._discretizeStates = False
        self._discretizeActions = False

        self.native = self

        # if proxy:
        #     native = proxy
        # self.native = native if native else self

        # Indexing
        self.index = Entity.number
        self.indexKind = Entity.indexes.get(self.kind, 0)
        Entity.number += 1
        Entity.indexes[self.kind] = self.indexKind + 1

        self._properties = {}
        self._children = []
        self.parent = None
        self.activated = False

        self.filterObservables = None

    def _sid(self, serializer, raw=False):
        if self.absoluteName:
            sid = f'#{self.absoluteName}'
        else:
            sid = f'{self.kind}:{self.indexKind}'
        if raw:
            return sid
        else:
            return serializer.uid('entity', sid)

    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['kind', 'absoluteName', 'index', 'indexKind', 'parent', '_children', '_properties'])
        return dict_

    # Links and proxies
    def linkedTo(self, entity):
        return self.native == entity.native

    def convertTo(self, manager=None):
        manager = parameter(manager, self.manager)
        return manager.convertEntity(self)
    
    def createLinkedEntity(self, manager, proxy=True):
        if self.parent:
            parent = self.parent.convertTo(manager)
        entity = manager.proxyCls(self, manager=manager)
        if self.parent:
            parent.addChild(entity)
        return entity
    
    # Children
    def addChild(self, entity):
        if entity not in self._children:
            if entity.absoluteName and self.findAbsoluteName(entity.absoluteName):
                raise Exception(f'An entity/property named {entity.absoluteName} already exists within {self.root.reference()}! \
                                 Names should be uniques.')
            entity.parent = self
            self._children.append(entity)
            # if entity.PHYSICAL:
            #     self.physicals.append(entity)
            if self.activated:
                entity.activate()

    def removeChild(self, entity):
        if entity in self._children:
            #del self._children[entity.name]
            if self.activated:
                entity.deactivate()
            self._children.remove(entity)
            # if entity.PHYSICAL:
            #     self.physicals.remove(entity)
            entity.parent = None

    def clearChildren(self):
        entities = list(self._children)
        for entity in entities:
            self.removeChild(entity)

    @property
    def root(self):
        if self.parent:
            return self.parent.root
        return self

    @property
    def manager(self):
        return self.root._manager

    def isRoot(self):
        return not self.parent and self.kind == 'root'
    
    def child(self, filter_=None):
        # self.findAbsoluteName(name, fromRoot=fromRoot, onlyEntities=True)
        return (self.children(filter_) + [None])[0]

    def children(self, filter_=None):
        if not filter_:
            return list(self._children)

        filtered = self._filterChild(filter_)
        return list(filter(filtered, self._children))

    def cascadingChild(self, filter_=None):
        children = self.children(filter_)
        return (children + [child for entity in children for child in entity.cascadingChildren(filter_)] + [None])[0]

    def cascadingChildren(self, filter_=None):
        children = self.children(filter_)
        return children + [child for entity in children for child in entity.cascadingChildren(filter_)]

    def _filterChild(self, filter_):
        if not filter_:
            filtered = lambda child: True
        if filter_[0] == '#':
            filtered = lambda child: child.absoluteName == filter_[1:]
        elif ':' in filter_:
            kind, index = filter_.split(':')
            filtered = lambda child: child.kind == kind and child.indexKind == index
        else:
            filtered = lambda child: child.kind == filter_
        return filtered

    # Properties and children
    def findAbsoluteName(self, name, fromRoot=True, onlyEntities=False):
        if not name:
            return None
        children = (self.root if fromRoot else self).cascadingChildren()
        for child in children:
            if child.absoluteName == name:
                return child
            if not onlyEntities:
                properties = child.cascadingProperties()
                for prop in properties:
                    if prop.absoluteName == name:
                        return prop
        return None
    
    # Properties
    @property
    def _proxifiedProperties(self):
        return self._properties

    def addProperty(self, prop):
        self._properties[prop.name] = prop

    def removeProperty(self, prop):
        if prop.name in self._properties:
            del self._properties[prop.name]

    def propertyItem(self, filter_=None):
        return (self.properties(filter_) + [None])[0]

    def properties(self, filter_=None, visual=None):
        if not filter_:
            properties = list(self._proxifiedProperties.values())
        else:
            # Omit first dot
            if filter_[0] == '.':
                filter_ = filter_[1:]
            if filter_[0] == '#':
                def filtered(property):
                    return property.absoluteName == filter_[1:]
            else:
                def filtered(property):
                    return property.name == filter_

            properties = list(filter(filtered, self._proxifiedProperties.values()))
        if visual is not None:
            properties = [prop for prop in properties if prop.visual == visual]
        return properties

    def cascadingProperties(self, filter_=None, visual=None):
        filterChildren, filterProperties = None, None
        if filter_:
            filterChildren, filterProperties = (
                filter_.split('.') + [None])[:2]
        properties = []
        if not filterChildren or self._filterChild(filterChildren)(self):
            properties = self.properties(filterProperties, visual=visual)
        return properties + \
            [property for entity in self.children(
                filterChildren) for property in entity.cascadingProperties(filter_, visual=visual)]

    def cascadingProperty(self, filter_=None):
        return (self.cascadingProperties(filter_) + [None])[0]
    
    def visualProperties(self):
        return self.properties(visual=True)

    def cascadingVisualProperties(self):
        return self.cascadingProperties(visual=True)
    
    def visualPropertyNames(self):
        properties = self.root.cascadingVisualProperties()
        return set([(prop.name, prop.dim) for prop in properties])

    def observeVisualProperties(self):
        values = []
        for name, dim in self.visualPropertyNames():
            prop = self.propertyItem(f'.{name}')
            if prop:
                values += prop.observePlain().tolist()
            else:
                values += [0] * dim
        return values
    
    def observe(self, spaces=None, formatParameters=None):
        spaces = spaces.convertTo(
            self.manager, kind=SpaceKind.BASIC) if spaces else None
        if self.filterObservables:
            obsSpaces = [x.space for x in self.filterObservables]
            if not spaces:
                spaces = obsSpaces
            else:
                spaces = [x for x in spaces if x in obsSpaces]

        ys = []
        # spaces = spaces if spaces else self.outcomeSpaces
        for property_ in self.cascadingProperties():
            if property_.observable() and property_.bounded:
                y = property_.observe(formatParameters=formatParameters)
                if spaces is None or y.space in spaces:
                    ys.append(y)
        return Observation(*ys)

    def observeFrom(self, spaces=None, formatParameters=None):
        return self.world.observe(spaces=spaces, formatParameters=formatParameters)

    def currentContext(self, dataset=None, spaces=None):
        obs = self.observe(spaces=spaces)
        if dataset:
            obs = obs.convertTo(dataset)
        return obs
    
    # Routines
    def activate(self):
        if self.activated:
            return
        self.activated = True

        self._activate()
        for e in self._children:
            e.activate()

        for p in self._properties.values():
            p._activate()

    def deactivate(self):
        if not self.activated:
            return
        self.activated = False

        self._deactivate()
        for e in self._children:
            e.deactivate()

    def _activate(self):
        pass

    def _deactivate(self):
        pass

    # Discretization
    @property
    def discretizeStates(self):
        return self.root._discretizeStates

    @discretizeStates.setter
    def discretizeStates(self, discretizeStates):
        if discretizeStates == self._discretizeStates:
            return
        assert(self.root.CAN_BE_DISCRETIZED)
        self._discretizeStates = discretizeStates
        self._discretizationChanged(False, discretizeStates)

    @property
    def discretizeActions(self):
        return self.root._discretizeActions

    @discretizeActions.setter
    def discretizeActions(self, discretizeActions):
        if discretizeActions == self._discretizeActions:
            return
        assert(self.root.CAN_BE_DISCRETIZED)
        self._discretizeActions = discretizeActions
        self._discretizationChanged(True, discretizeActions)

    def _discretizationChanged(self, action, discrete):
        pass

    @property
    def discreteStates(self):
        return self.discretizeStates or self.root.DISCRETE_STATES

    @property
    def discreteActions(self):
        return self.discretizeActions or self.root.DISCRETE_ACTIONS

    @property
    def world(self):
        # Supposed to be the root
        return self.root

    @property
    def engine(self):
        return self.manager.engine

    def observables(self):
        return [p for p in self.cascadingProperties() if p.observable()]

    def actions(self):
        actions = [p for p in self.cascadingProperties() if p.controllable()]
        if self.discretizeActions:
            actions = {(p, name): p.space.point(params)
                       for p in actions for name, params in p.actions.items()}
        return Ensemble(actions)
    
    # String
    def fullname(self):
        return self.parent.fullname() + "." + self.absoluteName if self.parent else ""

    def reference(self, short=False):
        if short:
            if self.absoluteName:
                s = f'#{self.absoluteName}'
            else:
                s = f'{self.kind}:{self.indexKind}'
        else:
            s = f'{self.kind}:{self.indexKind}'
            if self.absoluteName:
                s += f'#{self.absoluteName}'
        return s

    def __repr__(self):
        s = f'{self.kind}:{self.indexKind}'
        if self.disconnected:
            s = colorText('❌', Colors.RED) + s
        if self.absoluteName:
            s += f'#{self.absoluteName}'
        if self.parent:
            s += f' bound to {self.parent.reference()}'
        return s
    

class ProxyEntity(Entity):
    def __init__(self, entity, manager=None):
        manager = parameter(manager, entity.manager)
        super().__init__(entity.kind, entity.absoluteName,
                         manager=manager)

        self.proxy = entity
        self.native = entity.native
    
    @property
    def _proxifiedProperties(self):
        _properties = dict(self.proxy._properties)
        _properties.update(self._properties)
        return _properties




import numpy as np
import random
import copy
import math

import time

# import matplotlib.pyplot as plt
# from mpl_toolkits.mplot3d import Axes3D

from ..learner import Learner

from exlab.utils.io import parameter
from dinos.utils.move import MoveConfig
# from ....utils.maths import uniformSampling, iterrange

# from ....data.data import InteractionEvent
# from ....data.dataset import Dataset
# from ....models.regression import RegressionModel
# from ....planners.planner import Planner
from dinos.agents.tools.datasets.model_dataset import ModelDataset
from dinos.agents.tools.models.regression import RegressionModel


class ModelLearner(Learner):
    MODEL_CLASS = RegressionModel
    DATASET_CLASS = ModelDataset
    MULTI_EPISODE = 1

    def __init__(self, environment, dataset=None, performer=None, planner=None, options={}):
        dataset = parameter(dataset, self.DATASET_CLASS(modelClass=self.MODEL_CLASS))
        super().__init__(environment, dataset, performer, planner, options)

    # def _preEpisode(self):
    #     # Choose learning strategy randomly
    #     strategy = self.trainStrategies.sample()
    #     config = MoveConfig(strategy=strategy)

    #     self.logger.debug('Strategy used at iteration {}: {}'.format(
    #         self.iteration, config.strategy), 'STRAT')
    #     return config

    # def _postEpisode(self, memory):
    #     super()._postEpisode(memory)
